from __future__ import annotations

import logging
from fractions import Fraction
from typing import Any, Union

import numpy as np
from numpy.random import Generator

from newclid.jgex.angles import ang_between, ang_of
from newclid.jgex.geometries import (
    LENGTH_UNIT,
    JGEXCircle,
    JGEXGeometry,
    JGEXLine,
    JGEXPoint,
    circle_circle_intersection,
    line_circle_intersection,
    line_line_intersection,
)
from newclid.numerical import close_enough, nearly_zero
from newclid.tools import str_to_fraction

LOGGER = logging.getLogger(__name__)


def sketch(
    name: str, args: tuple[Union[JGEXPoint, str], ...], rng: Generator
) -> list[JGEXGeometry]:
    fun = globals()["sketch_" + name]
    res = fun(args, rng=rng)
    if isinstance(res, list) or isinstance(res, tuple):
        return list(res)  # type: ignore
    return [res]


def sketch_function_name(name: str) -> str:
    return name.split("sketch_")[1]


def sketch_aline(args: tuple[JGEXPoint, ...], **kwargs: Any) -> JGEXLine:
    """Sketch the construction aline."""
    A, B, C, D, E = args
    ab = A - B
    cb = C - B
    de = D - E

    ang_ab = np.arctan2(ab.y, ab.x)
    ang_bc = np.arctan2(cb.y, cb.x)
    ang_de = np.arctan2(de.y, de.x)

    ang_ex = ang_de + ang_bc - ang_ab
    X = E + LENGTH_UNIT * JGEXPoint(x=np.cos(ang_ex), y=np.sin(ang_ex))
    return JGEXLine(E, X)


def sketch_aline0(args: tuple[JGEXPoint, ...], **kwargs: Any) -> JGEXLine:
    """Sketch the construction aline0."""
    A, B, C, D, E, F, G = args
    ab = A - B
    cd = C - D
    ef = E - F

    ang_ab = np.arctan2(ab.y, ab.x)
    ang_cd = np.arctan2(cd.y, cd.x)
    ang_ef = np.arctan2(ef.y, ef.x)

    ang_ex = ang_ef + ang_cd - ang_ab
    X = G + LENGTH_UNIT * JGEXPoint(x=np.cos(ang_ex), y=np.sin(ang_ex))
    return JGEXLine(G, X)


def sketch_acircle(args: tuple[JGEXPoint, ...], **kwargs: Any) -> JGEXCircle:
    a, b, c, d, f = args
    de = sketch_aline((c, a, b, f, d))
    fe = sketch_aline((a, c, b, d, f))
    (e,) = line_line_intersection(de, fe)
    return JGEXCircle(p1=d, p2=e, p3=f)


def sketch_amirror(
    args: tuple[JGEXPoint, JGEXPoint, JGEXPoint], **kwargs: Any
) -> JGEXLine:
    """Sketch the angle mirror."""
    A, B, C = args
    ab = A - B
    cb = C - B

    dab = A.distance(B)
    ang_ab = np.arctan2(ab.y / dab, ab.x / dab)
    dcb = C.distance(B)
    ang_bc = np.arctan2(cb.y / dcb, cb.x / dcb)

    ang_bx = 2 * ang_bc - ang_ab
    X = B + LENGTH_UNIT * JGEXPoint(x=np.cos(ang_bx), y=np.sin(ang_bx))
    return JGEXLine(B, X)


def sketch_bisect(
    args: tuple[JGEXPoint, JGEXPoint, JGEXPoint], **kwargs: Any
) -> JGEXLine:
    a, b, c = args
    ab = a.distance(b)
    bc = b.distance(c)
    x = b + (c - b) * (ab / bc)
    m = (a + x) / 2
    return JGEXLine(b, m)


def sketch_exbisect(
    args: tuple[JGEXPoint, JGEXPoint, JGEXPoint], **kwargs: Any
) -> JGEXLine:
    _, b, _ = args
    return sketch_bisect(args).perpendicular_line(b)


def sketch_bline(args: tuple[JGEXPoint, JGEXPoint], **kwargs: Any) -> JGEXLine:
    a, b = args
    m = (a + b) / 2
    return m.perpendicular_line(JGEXLine(a, b))


def sketch_dia(args: tuple[JGEXPoint, JGEXPoint], **kwargs: Any) -> JGEXCircle:
    a, b = args
    return JGEXCircle((a + b) / 2, p1=a)


def sketch_tangent(
    args: tuple[JGEXPoint, JGEXPoint, JGEXPoint], **kwargs: Any
) -> tuple[JGEXPoint, ...]:
    a, o, b = args
    dia = sketch_dia((a, o))
    return circle_circle_intersection(JGEXCircle(o, p1=b), dia)


def sketch_circle(
    args: tuple[JGEXPoint, JGEXPoint, JGEXPoint], **kwargs: Any
) -> JGEXCircle:
    a, b, c = args
    return JGEXCircle(center=a, radius=b.distance(c))


def sketch_cc_tangent(
    args: tuple[JGEXPoint, JGEXPoint, JGEXPoint, JGEXPoint], **kwargs: Any
) -> tuple[JGEXPoint, ...]:
    """Sketch external tangents to two circles."""
    o, a, w, b = args
    ra, rb = o.distance(a), w.distance(b)

    ow = JGEXLine(o, w)

    if close_enough(ra, rb):
        oo = ow.perpendicular_line(o)
        oa = JGEXCircle(o, ra)
        x, z = line_circle_intersection(oo, oa, expected_points=2)
        y = x + w - o
        t = z + w - o
        return x, y, z, t

    swap = rb > ra
    if swap:
        o, a, w, b = w, b, o, a
        ra, rb = rb, ra

    if close_enough(o.distance(w), abs(rb - ra)):
        LOGGER.debug(
            "The circles are internally tangent. Try another definition of a tangent line."
        )
    elif o.distance(w) < abs(rb - ra):
        LOGGER.debug("One circle is inside the other and there are no tangents.")

    oa = JGEXCircle(o, ra)
    q = o + (w - o) * ra / (ra - rb)

    x, z = circle_circle_intersection(sketch_dia((o, q)), oa, expected_points=2)
    y = w.foot(JGEXLine(x, q))
    i = w.foot(JGEXLine(z, q))

    if swap:
        x, y, z, i = y, x, i, z

    return x, y, z, i


def sketch_e5128(
    args: tuple[JGEXPoint, ...], **kwargs: Any
) -> tuple[JGEXPoint, JGEXPoint]:
    a, b, c, d = args

    g = (a + b) / 2
    de = JGEXLine(d, g)

    e, f = line_circle_intersection(de, JGEXCircle(c, p1=b), expected_points=2)

    if e.distance(d) < f.distance(d):
        e = f
    return e, g


def random_rfss(*points: JGEXPoint, rng: Generator) -> list[JGEXPoint]:
    """Random rotate-flip-scale-shift a JGEXPoint cloud."""
    # center point cloud.
    average = sum(points, JGEXPoint(x=0.0, y=0.0)) / len(points)
    points = tuple(p - average for p in points)

    # rotate
    ang = rng.uniform(0.0, 2 * np.pi)
    sin, cos = np.sin(ang), np.cos(ang)
    # scale and shift
    scale = rng.uniform(0.5, 2.0)
    shift = LENGTH_UNIT * JGEXPoint(x=rng.uniform(-1, 1), y=rng.uniform(-1, 1))
    points = tuple(p.rotate(sin, cos) * scale + shift for p in points)

    # randomly flip
    if rng.random() < 0.5:
        points = tuple(p.flip() for p in points)

    return list(points)


def head_from(tail: JGEXPoint, ang: float, length: float = LENGTH_UNIT) -> JGEXPoint:
    vector = JGEXPoint(x=np.cos(ang) * length, y=np.sin(ang) * length)
    return tail + vector


def sketch_eq_quadrangle(
    args: tuple[JGEXPoint, ...], rng: Generator
) -> tuple[JGEXPoint, ...]:
    """Sketch quadrangle with two equal opposite sides."""
    a = LENGTH_UNIT * JGEXPoint(x=0.0, y=0.0)
    b = LENGTH_UNIT * JGEXPoint(x=1.0, y=0.0)

    length = LENGTH_UNIT * rng.uniform(0.5, 2.0)
    ang = rng.uniform(np.pi / 3, np.pi * 2 / 3)
    d = head_from(a, ang, length)

    ang = ang_of(b, d)
    ang = rng.uniform(ang / 10, ang / 9)
    c = head_from(b, ang, length)
    a, b, c, d = random_rfss(a, b, c, d, rng=rng)
    return a, b, c, d


def sketch_iso_trapezoid(
    args: tuple[JGEXPoint, ...], rng: Generator
) -> tuple[JGEXPoint, ...]:
    a = LENGTH_UNIT * JGEXPoint(x=0.0, y=0.0)
    b = LENGTH_UNIT * JGEXPoint(x=1.0, y=0.0)
    rel_lenght = rng.uniform(0.5, 2.0)
    rel_height = rng.uniform(0.5, 2.0)
    c = LENGTH_UNIT * JGEXPoint(x=0.5 + rel_lenght / 2.0, y=rel_height)
    d = LENGTH_UNIT * JGEXPoint(x=0.5 - rel_lenght / 2.0, y=rel_height)
    a, b, c, d = random_rfss(a, b, c, d, rng=rng)
    return a, b, c, d


def sketch_eqangle2(args: tuple[JGEXPoint, ...], rng: Generator) -> JGEXPoint:
    """Sketch the def eqangle2."""
    a, b, c = args

    ba = b.distance(a)
    bc = b.distance(c)
    length = ba * ba / bc

    if rng.uniform(0.0, 1.0) < 0.5:
        be = min(length, bc)
        be = rng.uniform(be * 0.1, be * 0.9)
    else:
        be = max(length, bc)
        be = rng.uniform(be * 1.1, be * 1.5)

    e = b + (c - b) * (be / bc)
    y = b + (a - b) * (be / length)
    return line_line_intersection(JGEXLine(c, y), JGEXLine(a, e))[0]


def sketch_eqangle3(args: tuple[JGEXPoint, ...], **kwargs: Any) -> JGEXCircle:
    a, b, d, e, f = args
    de = d.distance(e)
    ef = e.distance(f)
    ab = b.distance(a)
    ang_ax = ang_of(a, b) + ang_between(e, f, d)
    x = head_from(a, ang_ax, length=de / ef * ab)
    return JGEXCircle(p1=a, p2=b, p3=x)


def sketch_eqdia_quadrangle(
    args: tuple[JGEXPoint, ...], rng: Generator
) -> tuple[JGEXPoint, ...]:
    """Sketch quadrangle with two equal diagonals."""
    rel_x = rng.uniform(0.3, 0.7)
    rel_y = rng.uniform(0.3, 0.7)
    a = LENGTH_UNIT * JGEXPoint(x=-rel_x, y=0.0)
    c = LENGTH_UNIT * JGEXPoint(x=1 - rel_x, y=0.0)
    b = LENGTH_UNIT * JGEXPoint(x=0.0, y=-rel_y)
    d = LENGTH_UNIT * JGEXPoint(x=0.0, y=1 - rel_y)

    ang = rng.uniform(-0.25 * np.pi, 0.25 * np.pi)
    sin, cos = np.sin(ang), np.cos(ang)
    b = b.rotate(sin, cos)
    d = d.rotate(sin, cos)
    a, b, c, d = random_rfss(a, b, c, d, rng=rng)
    return a, b, c, d


def random_points(n: int, rng: Generator) -> list[JGEXPoint]:
    return [
        LENGTH_UNIT * JGEXPoint(x=rng.uniform(-1, 1), y=rng.uniform(-1, 1))
        for _ in range(n)
    ]


def sketch_free(args: tuple[JGEXPoint, ...], rng: Generator) -> JGEXPoint:
    return random_points(1, rng)[0]


def sketch_isos(args: tuple[JGEXPoint, ...], rng: Generator) -> tuple[JGEXPoint, ...]:
    rel_base = rng.uniform(0.5, 1.5)
    rel_height = rng.uniform(0.5, 1.5)

    b = LENGTH_UNIT * JGEXPoint(x=-rel_base / 2, y=0.0)
    c = LENGTH_UNIT * JGEXPoint(x=rel_base / 2, y=0.0)
    a = LENGTH_UNIT * JGEXPoint(x=0.0, y=rel_height)
    a, b, c = random_rfss(a, b, c, rng=rng)
    return a, b, c


def sketch_line(args: tuple[JGEXPoint, ...], **kwargs: Any) -> JGEXLine:
    a, b = args
    return JGEXLine(a, b)


def sketch_cyclic(args: tuple[JGEXPoint, ...], **kwargs: Any) -> JGEXCircle:
    a, b, c = args
    return JGEXCircle(p1=a, p2=b, p3=c)


def sketch_midp(args: tuple[JGEXPoint, ...], **kwargs: Any) -> JGEXPoint:
    a, b = args
    return (a + b) / 2


def sketch_pentagon(
    args: tuple[JGEXPoint, ...], rng: Generator
) -> tuple[JGEXPoint, ...]:
    points = [LENGTH_UNIT * JGEXPoint(x=1.0, y=0.0)]
    ang = 0.0

    for i in range(4):
        ang += (2 * np.pi - ang) / (5 - i) * rng.uniform(0.5, 1.5)
        point = LENGTH_UNIT * JGEXPoint(x=np.cos(ang), y=np.sin(ang))
        points.append(point)

    a, b, c, d, e = points
    a, b, c, d, e = random_rfss(a, b, c, d, e, rng=rng)
    return a, b, c, d, e


def sketch_pline(args: tuple[JGEXPoint, ...], **kwargs: Any) -> JGEXLine:
    a, b, c = args
    return a.parallel_line(JGEXLine(b, c))


def sketch_pmirror(args: tuple[JGEXPoint, ...], **kwargs: Any) -> JGEXPoint:
    a, b = args
    return b * 2 - a


def sketch_quadrangle(
    args: tuple[JGEXPoint, ...], rng: Generator
) -> tuple[JGEXPoint, ...]:
    """Sketch a random quadrangle."""
    rel_x = rng.uniform(0.3, 0.7)
    a = LENGTH_UNIT * JGEXPoint(x=-rel_x, y=0.0)
    c = LENGTH_UNIT * JGEXPoint(x=1 - rel_x, y=0.0)
    b = LENGTH_UNIT * JGEXPoint(x=0.0, y=-rng.uniform(0.25, 0.75))
    d = LENGTH_UNIT * JGEXPoint(x=0.0, y=rng.uniform(0.25, 0.75))

    ang = rng.uniform(-0.25 * np.pi, 0.25 * np.pi)
    sin, cos = np.sin(ang), np.cos(ang)
    b = b.rotate(sin, cos)
    d = d.rotate(sin, cos)
    a, b, c, d = random_rfss(a, b, c, d, rng=rng)
    return a, b, c, d


def sketch_r_trapezoid(
    args: tuple[JGEXPoint, ...], rng: Generator
) -> tuple[JGEXPoint, ...]:
    a = LENGTH_UNIT * JGEXPoint(x=0.0, y=1.0)
    d = LENGTH_UNIT * JGEXPoint(x=0.0, y=0.0)
    b = LENGTH_UNIT * JGEXPoint(x=rng.uniform(0.5, 1.5), y=1.0)
    c = LENGTH_UNIT * JGEXPoint(x=rng.uniform(0.5, 1.5), y=0.0)
    a, b, c, d = random_rfss(a, b, c, d, rng=rng)
    return a, b, c, d


def sketch_r_triangle(
    args: tuple[JGEXPoint, ...], rng: Generator
) -> tuple[JGEXPoint, ...]:
    a = LENGTH_UNIT * JGEXPoint(x=0.0, y=0.0)
    b = LENGTH_UNIT * JGEXPoint(x=0.0, y=rng.uniform(0.5, 2.0))
    c = LENGTH_UNIT * JGEXPoint(x=rng.uniform(0.5, 2.0), y=0.0)
    a, b, c = random_rfss(a, b, c, rng=rng)
    return a, b, c


def sketch_rectangle(
    args: tuple[JGEXPoint, ...], rng: Generator
) -> tuple[JGEXPoint, ...]:
    a = LENGTH_UNIT * JGEXPoint(x=0.0, y=0.0)
    b = LENGTH_UNIT * JGEXPoint(x=0.0, y=1.0)
    rel_lenght = rng.uniform(0.5, 2.0)
    c = LENGTH_UNIT * JGEXPoint(x=rel_lenght, y=1.0)
    d = LENGTH_UNIT * JGEXPoint(x=rel_lenght, y=0.0)
    a, b, c, d = random_rfss(a, b, c, d, rng=rng)
    return a, b, c, d


def sketch_reflect(args: tuple[JGEXPoint, ...], **kwargs: Any) -> JGEXPoint:
    a, b, c = args
    m = a.foot(JGEXLine(b, c))
    return m * 2 - a


def sketch_risos(args: tuple[JGEXPoint, ...], rng: Generator) -> tuple[JGEXPoint, ...]:
    a = LENGTH_UNIT * JGEXPoint(x=0.0, y=0.0)
    b = LENGTH_UNIT * JGEXPoint(x=0.0, y=1.0)
    c = LENGTH_UNIT * JGEXPoint(x=1.0, y=0.0)
    a, b, c = random_rfss(a, b, c, rng=rng)
    return a, b, c


def sketch_rotaten90(args: tuple[JGEXPoint, ...], **kwargs: Any) -> JGEXPoint:
    a, b = args
    ang = -np.pi / 2
    return a + (b - a).rotate(np.sin(ang), np.cos(ang))


def sketch_rotatep90(args: tuple[JGEXPoint, ...], **kwargs: Any) -> JGEXPoint:
    a, b = args
    ang = np.pi / 2
    return a + (b - a).rotate(np.sin(ang), np.cos(ang))


def sketch_s_angle(args: tuple[JGEXPoint, JGEXPoint, str], **kwargs: Any) -> JGEXLine:
    a, b, angle = args
    f = str_to_fraction(angle)
    ang = float(f) * np.pi
    x = b + (a - b).rotatea(ang)
    return JGEXLine(b, x)


def sketch_aconst(
    args: tuple[JGEXPoint, JGEXPoint, JGEXPoint, str], **kwargs: Any
) -> JGEXLine:
    a, b, c, angle = args
    f = str_to_fraction(angle)
    ang = float(f) * np.pi
    x = c + (a - b).rotatea(ang)
    return JGEXLine(c, x)


def sketch_segment(
    args: tuple[JGEXPoint, ...], rng: Generator
) -> tuple[JGEXPoint, JGEXPoint]:
    a, b = random_points(2, rng)
    return a, b


def sketch_shift(args: tuple[JGEXPoint, ...], **kwargs: Any) -> JGEXPoint:
    a, b, c = args
    return c + (b - a)


def sketch_square(
    args: tuple[JGEXPoint, ...], **kwargs: Any
) -> tuple[JGEXPoint, JGEXPoint]:
    a, b = args
    c = b + (a - b).rotatea(-np.pi / 2)
    d = a + (b - a).rotatea(np.pi / 2)
    return c, d


def sketch_isquare(
    args: tuple[JGEXPoint, ...], rng: Generator
) -> tuple[JGEXPoint, ...]:
    a = LENGTH_UNIT * JGEXPoint(x=0.0, y=0.0)
    b = LENGTH_UNIT * JGEXPoint(x=1.0, y=0.0)
    c = LENGTH_UNIT * JGEXPoint(x=1.0, y=1.0)
    d = LENGTH_UNIT * JGEXPoint(x=0.0, y=1.0)
    a, b, c, d = random_rfss(a, b, c, d, rng=rng)
    return a, b, c, d


def sketch_tline(args: tuple[JGEXPoint, ...], **kwargs: Any) -> JGEXLine:
    a, b, c = args
    return a.perpendicular_line(JGEXLine(b, c))


def sketch_trapezoid(
    args: tuple[JGEXPoint, ...], rng: Generator
) -> tuple[JGEXPoint, ...]:
    d = LENGTH_UNIT * JGEXPoint(x=0.0, y=0.0)
    c = LENGTH_UNIT * JGEXPoint(x=1.0, y=0.0)

    base = LENGTH_UNIT * rng.uniform(0.5, 2.0)
    height = LENGTH_UNIT * rng.uniform(0.5, 2.0)
    a = JGEXPoint(x=LENGTH_UNIT * rng.uniform(-0.5, 1.5), y=height)
    b = JGEXPoint(x=a.x + base, y=height)
    a, b, c, d = random_rfss(a, b, c, d, rng=rng)
    return a, b, c, d


def sketch_triangle(
    args: tuple[JGEXPoint, ...], rng: Generator
) -> tuple[JGEXPoint, JGEXPoint, JGEXPoint]:
    a = LENGTH_UNIT * JGEXPoint(x=0.0, y=0.0)
    b = LENGTH_UNIT * JGEXPoint(x=1.0, y=0.0)
    ac = LENGTH_UNIT * rng.uniform(0.5, 2.0)
    ang = rng.uniform(0.2, 0.8) * np.pi
    c = head_from(a, ang, ac)
    a, b, c = random_rfss(a, b, c, rng=rng)
    return a, b, c


def sketch_triangle12(
    args: tuple[JGEXPoint, ...], rng: Generator
) -> tuple[JGEXPoint, JGEXPoint, JGEXPoint]:
    b = LENGTH_UNIT * JGEXPoint(x=0.0, y=0.0)
    c = LENGTH_UNIT * JGEXPoint(x=rng.uniform(1.5, 2.5), y=0.0)

    circle_1 = JGEXCircle(b, LENGTH_UNIT)
    circle_2 = JGEXCircle(c, 2.0 * LENGTH_UNIT)
    a, _ = circle_circle_intersection(circle_1, circle_2, expected_points=2)
    a, b, c = random_rfss(a, b, c, rng=rng)
    return a, b, c


def sketch_trisect(
    args: tuple[JGEXPoint, ...], **kwargs: Any
) -> tuple[JGEXPoint, JGEXPoint]:
    """Sketch two trisectors of an angle."""
    a, b, c = args
    ang1 = ang_of(b, a)
    ang2 = ang_of(b, c)

    swap = 0
    if ang1 > ang2:
        ang1, ang2 = ang2, ang1
        swap += 1

    if ang2 - ang1 > np.pi:
        ang1, ang2 = ang2, ang1 + 2 * np.pi
        swap += 1

    angx = ang1 + (ang2 - ang1) / 3
    angy = ang2 - (ang2 - ang1) / 3

    x = b + LENGTH_UNIT * JGEXPoint(x=np.cos(angx), y=np.sin(angx))
    y = b + LENGTH_UNIT * JGEXPoint(x=np.cos(angy), y=np.sin(angy))

    ac = JGEXLine(a, c)
    (x,) = line_line_intersection(JGEXLine(b, x), ac)
    (y,) = line_line_intersection(JGEXLine(b, y), ac)

    if swap == 1:
        return y, x
    return x, y


def sketch_trisegment(
    args: tuple[JGEXPoint, ...], **kwargs: Any
) -> tuple[JGEXPoint, JGEXPoint]:
    a, b = args
    x, y = a + (b - a) * (1.0 / 3), a + (b - a) * (2.0 / 3)
    return x, y


def sketch_ieq_triangle(
    args: tuple[JGEXPoint, ...], rng: Generator
) -> tuple[JGEXPoint, ...]:
    a = LENGTH_UNIT * JGEXPoint(x=0.0, y=0.0)
    b = LENGTH_UNIT * JGEXPoint(x=1.0, y=0.0)

    c, _ = circle_circle_intersection(
        JGEXCircle(a, p1=b), JGEXCircle(b, p1=a), expected_points=2
    )
    a, b, c = random_rfss(a, b, c, rng=rng)
    return a, b, c


def sketch_incenter2(
    args: tuple[JGEXPoint, ...], **kwargs: Any
) -> tuple[JGEXPoint, JGEXPoint, JGEXPoint, JGEXPoint]:
    a, b, c = args
    l1 = sketch_bisect((b, a, c))
    l2 = sketch_bisect((a, b, c))
    (i,) = line_line_intersection(l1, l2)
    x = i.foot(JGEXLine(b, c))
    y = i.foot(JGEXLine(c, a))
    z = i.foot(JGEXLine(a, b))
    return x, y, z, i


def sketch_excenter2(
    args: tuple[JGEXPoint, ...], **kwargs: Any
) -> tuple[JGEXPoint, JGEXPoint, JGEXPoint, JGEXPoint]:
    a, b, c = args
    l1 = sketch_bisect((b, a, c))
    l2 = sketch_exbisect((a, b, c))
    (i,) = line_line_intersection(l1, l2)
    x = i.foot(JGEXLine(b, c))
    y = i.foot(JGEXLine(c, a))
    z = i.foot(JGEXLine(a, b))
    return x, y, z, i


def sketch_centroid(
    args: tuple[JGEXPoint, ...], **kwargs: Any
) -> tuple[JGEXPoint, JGEXPoint, JGEXPoint, JGEXPoint]:
    a, b, c = args
    x = (b + c) / 2
    y = (c + a) / 2
    z = (a + b) / 2
    (i,) = line_line_intersection(JGEXLine(a, x), JGEXLine(b, y))
    return x, y, z, i


def sketch_ninepoints(
    args: tuple[JGEXPoint, ...], **kwargs: Any
) -> tuple[JGEXPoint, JGEXPoint, JGEXPoint, JGEXPoint]:
    a, b, c = args
    x = (b + c) / 2
    y = (c + a) / 2
    z = (a + b) / 2
    return x, y, z, JGEXCircle(p1=x, p2=y, p3=z).center


def sketch_2l1c(
    args: tuple[JGEXPoint, ...], **kwargs: Any
) -> tuple[JGEXPoint, JGEXPoint, JGEXPoint, JGEXPoint]:
    """Sketch a circle touching two lines and another circle."""
    a, b, c, p = args
    bc, ac = JGEXLine(b, c), JGEXLine(a, c)
    circle = JGEXCircle(p, p1=a)

    d, d_ = line_circle_intersection(
        p.perpendicular_line(bc), circle, expected_points=2
    )
    if bc.diff_side(d_, a):
        d = d_

    e, e_ = line_circle_intersection(
        p.perpendicular_line(ac), circle, expected_points=2
    )
    if ac.diff_side(e_, b):
        e = e_

    df = d.perpendicular_line(JGEXLine(p, d))
    ef = e.perpendicular_line(JGEXLine(p, e))
    (f,) = line_line_intersection(df, ef)

    g, g_ = line_circle_intersection(JGEXLine(c, f), circle, expected_points=2)
    if bc.same_side(g_, a):
        g = g_

    b_ = c + (b - c) / b.distance(c)
    a_ = c + (a - c) / a.distance(c)
    m = (a_ + b_) / 2
    (x,) = line_line_intersection(JGEXLine(c, m), JGEXLine(p, g))
    return x.foot(ac), x.foot(bc), g, x


def sketch_3peq(args: tuple[JGEXPoint, ...], rng: Generator) -> tuple[JGEXPoint, ...]:
    a, b, c = args
    ab, _, ca = JGEXLine(a, b), JGEXLine(b, c), JGEXLine(c, a)

    z = b + (c - b) * rng.uniform(-0.5, 1.5)

    z_ = z * 2 - c
    ca_parallel_line = z_.parallel_line(ca)
    (x,) = line_line_intersection(ca_parallel_line, ab)
    y = z * 2 - x
    return x, y, z


# NEW FUNCTIONS FOR NEW DEFINITIONS ---- V. S.


def sketch_isosvertex(args: tuple[JGEXPoint, ...], **kwargs: Any) -> JGEXLine:
    b, c = args
    m = (b + c) / 2.0

    return m.perpendicular_line(JGEXLine(b, c))


def sketch_eqratio(args: tuple[JGEXPoint, ...], **kwargs: Any) -> JGEXCircle:
    A, B, C, D, E, F, G = args

    dab = A.distance(B)
    dcd = C.distance(D)
    d_ef = E.distance(F)

    dgx = d_ef * dcd / dab
    return JGEXCircle(center=G, radius=dgx)


def sketch_rconst(
    args: tuple[JGEXPoint, JGEXPoint, JGEXPoint, str], **kwargs: Any
) -> JGEXCircle:
    """Sketches point x such that ab/cx=r"""
    A, B, C, r = args
    dab = A.distance(B)
    length = float(dab / float(str_to_fraction(r)))
    return JGEXCircle(center=C, radius=length)


def sketch_eqratio6(
    args: tuple[JGEXPoint, ...], **kwargs: Any
) -> Union[JGEXCircle, JGEXLine]:
    """Sketches a point x such that ax/cx=ef/gh"""
    A, C, E, F, G, H = args
    d_ef = E.distance(F)
    dgh = G.distance(H)

    if dgh == d_ef:
        M = (A + C) * 0.5
        return M.perpendicular_line(JGEXLine(A, C))

    else:
        ratio = d_ef / dgh
        extremum_1 = (1 / (1 - ratio)) * (A - ratio * C)
        extremum_2 = (1 / (1 + ratio)) * (ratio * C + A)
        center = (extremum_1 + extremum_2) / 2
        radius = extremum_1.distance(extremum_2) / 2
        return JGEXCircle(center=center, radius=radius)


def sketch_lconst(args: tuple[JGEXPoint, str], **kwargs: Any) -> JGEXCircle:
    """Sketches point x such that x in at lenght l of point a"""
    a, length = args
    return JGEXCircle(center=a, radius=float(str_to_fraction(length)))


def sketch_rconst2(
    args: tuple[JGEXPoint, JGEXPoint, str], **kwargs: Any
) -> Union[JGEXCircle, JGEXLine]:
    """Sketches point x such that ax/bx=r"""
    A, B, r = args
    ratio = float(str_to_fraction(r))

    if ratio == float(Fraction(1, 1)):
        M = (A + B) / 2
        return M.perpendicular_line(JGEXLine(A, B))

    extremum_1 = (1 / (1 - ratio)) * (A - ratio * B)  # pyright: ignore
    extremum_2 = (1 / (1 + ratio)) * (ratio * B + A)
    center = (extremum_1 + extremum_2) / 2
    radius = extremum_1.distance(extremum_2) / 2
    return JGEXCircle(center=center, radius=radius)


def sketch_between(
    args: tuple[JGEXPoint, ...], rng: Generator
) -> tuple[JGEXPoint, JGEXPoint, JGEXPoint]:
    """Sketches a segment ab and a point c between them."""
    a, b = random_points(2, rng)
    c = a + (b - a) * rng.uniform(0.0, 1.0)
    return c, a, b


def sketch_between_bound(
    args: tuple[JGEXPoint, JGEXPoint], rng: Generator
) -> JGEXPoint:
    """Sketches a point x between a and b."""
    a, b = args
    return a + (b - a) * rng.uniform(0.0, 1.0)


def sketch_iso_trapezoid2(
    args: tuple[JGEXPoint, JGEXPoint, JGEXPoint], rng: Generator
) -> JGEXPoint:
    a, b, c = args
    d1 = c.foot(JGEXLine(a, b))
    return c + (b - d1) + (a - d1)


def sketch_acute_triangle(
    args: tuple[JGEXPoint, ...], rng: Generator
) -> tuple[JGEXPoint, ...]:
    a, b = random_points(2, rng)
    # The locus of points such that c makes triangle a b c acute (with these specific a, b)
    # is such that (more or less) the x coordinate of c is between 0,1
    # and the y coordinate has to be above the circle with center .5,0
    # and radius .5 (if it is on the circle it is right, if it is below the circle then it is obtuse).
    t = rng.uniform(0, 1)
    length = a.distance(b)
    v = b - a
    height_min = np.sqrt(length**2 - ((t - 0.5) * length) ** 2)
    height = float(rng.uniform(height_min, 2.5 * height_min))  # pyright: ignore
    c = a + t * v + height * v.rot90()
    a, b, c = random_rfss(a, b, c, rng=rng)
    return a, b, c


def sketch_cc_itangent(
    args: tuple[JGEXPoint, JGEXPoint, JGEXPoint, JGEXPoint], **kwargs: Any
) -> tuple[JGEXPoint, ...]:
    """Sketch internal tangents to two circles."""
    o, a, w, b = args
    ra, rb = o.distance(a), w.distance(b)
    distance = o.distance(w)

    if close_enough(distance, abs(rb + ra)):
        LOGGER.info(
            "The circles are externally tangent. Try another definition of a tangent line."
        )
    elif distance < abs(rb + ra):
        LOGGER.info(
            "The circles intersect or one is inside the other. There are no internal tangents"
        )

    oa = JGEXCircle(o, ra)
    q = o + (w - o) * ra / (ra + rb)

    x, z = circle_circle_intersection(sketch_dia((o, q)), oa, expected_points=2)
    y = w.foot(JGEXLine(x, q))
    i = w.foot(JGEXLine(z, q))

    return x, y, z, i


def sketch_simtri(args: tuple[JGEXPoint, ...], **kwargs: Any) -> JGEXPoint:
    """Sketch the construction of the third vertex of a similar triangle."""
    A, B, C, P, Q = args
    ab = B - A
    ac = C - A
    pq = Q - P

    ang_ab = np.arctan2(ab.y, ab.x)
    ang_ac = np.arctan2(ac.y, ac.x)
    ang_pq = np.arctan2(pq.y, pq.x)
    ratio = A.distance(B) / P.distance(Q)
    length = A.distance(C) / ratio

    ang_pr = ang_pq + ang_ac - ang_ab
    r = P + length * JGEXPoint(x=np.cos(ang_pr), y=np.sin(ang_pr))
    return r


def sketch_simtrir(args: tuple[JGEXPoint, ...], **kwargs: Any) -> JGEXPoint:
    """Sketch the construction of the third vertex of a similar triangle with a reflection."""
    A, B, C, P, Q = args
    ab = B - A
    ac = C - A
    pq = Q - P

    ang_ab = np.arctan2(ab.y, ab.x)
    ang_ac = np.arctan2(ac.y, ac.x)
    ang_pq = np.arctan2(pq.y, pq.x)
    ratio = A.distance(B) / P.distance(Q)
    length = A.distance(C) / ratio

    ang_pr = ang_pq + ang_ac - ang_ab
    r1 = P + length * JGEXPoint(x=np.cos(ang_pr), y=np.sin(ang_pr))

    r, r2 = circle_circle_intersection(
        JGEXCircle(center=P, p1=r1), JGEXCircle(center=Q, p1=r1), expected_points=2
    )

    if nearly_zero(r.distance(r1)):
        r = r2

    return r


def sketch_contri(
    args: tuple[JGEXPoint, ...], rng: Generator, **kwargs: Any
) -> tuple[JGEXPoint, ...]:
    """Sketch the construction of the vertices of a congruent triangle."""
    A, B, C, P = args

    # rotation parameters
    ang = rng.uniform(0.0, 2 * np.pi)
    sin, cos = np.sin(ang), np.cos(ang)

    # take triangle to the origin
    q1 = B - A
    r1 = C - A

    # rotate the triangle
    q2 = q1.rotate(sin, cos)
    r2 = r1.rotate(sin, cos)

    # position triangle based on P
    q = q2 + P
    r = r2 + P

    return q, r


def sketch_contrir(
    args: tuple[JGEXPoint, ...], rng: Generator, **kwargs: Any
) -> tuple[JGEXPoint, ...]:
    """Sketch the construction of the vertices of a congruent triangle with a reflection."""
    A, B, C, P = args

    # rotation parameters
    ang = rng.uniform(0.0, 2 * np.pi)
    sin, cos = np.sin(ang), np.cos(ang)

    # take triangle to the origin
    q1 = B - A
    r1 = C - A

    # rotate the triangle
    q2 = q1.rotate(sin, cos)
    r2 = r1.rotate(sin, cos)

    # position triangle based on P
    q = q2 + P
    r3 = r2 + P

    r, r4 = circle_circle_intersection(
        JGEXCircle(center=P, p1=r3), JGEXCircle(center=q, p1=r3), expected_points=2
    )

    if nearly_zero(r.distance(r3)):
        r = r4

    return q, r


def sketch_test_r20(
    args: tuple[JGEXPoint, ...], rng: Generator
) -> tuple[JGEXPoint, ...]:
    b = JGEXPoint(x=0.0, y=0.0)
    a = JGEXPoint(x=0.0, y=rng.uniform(0.5, 2.0))
    c = JGEXPoint(x=rng.uniform(0.5, 2.0), y=0.0)
    o = (a + c) * 0.5
    o, a, b, c = random_rfss(o, a, b, c, rng=rng)
    return o, a, b, c


def sketch_test_r25(
    args: tuple[JGEXPoint, ...], rng: Generator
) -> tuple[JGEXPoint, ...]:
    a, b, c = random_points(3, rng)
    m = (a + b) * 0.5
    d = b + a - c
    return a, b, c, d, m


def sketch_l2const(args: tuple[JGEXPoint, str], **kwargs: Any) -> JGEXCircle:
    """Sketches point x such that x in at lenght sqrt(l) of point a"""

    a, l2 = args
    squared_length = float(str_to_fraction(l2))
    radius = squared_length**0.5
    return JGEXCircle(center=a, radius=radius)


def sketch_r2const(
    args: tuple[JGEXPoint, JGEXPoint, JGEXPoint, str], **kwargs: Any
) -> JGEXCircle:
    """Sketches point x such that (ab/cx)^2=r"""
    A, B, C, r = args
    dab = A.distance(B)
    length = float(dab / (float(str_to_fraction(r) ** 0.5)))
    return JGEXCircle(center=C, radius=length)



from __future__ import annotations

from typing import Generic, Self, TypeVar

from pydantic import BaseModel, model_validator

S = TypeVar("S")


class SymbolMergeHistory(BaseModel, Generic[S]):
    """Merge history for a symbol.

    Maintains a merge history to
    other Lines or Circles if they are (found out to be) equivalent

    Example:

        a -> b -
                \
            c -> d -> e -> f -> g


    d.merged_to = e
    representent(d) = g
    d.merged_from = {a, b, c, d}
    d.equivs = {a, b, c, d, e, f, g}

    """

    symbol: S
    fellows: set[S] = set()
    representent: SymbolMergeHistory[S] | None = None

    @model_validator(mode="after")
    def add_self_to_fellows(self) -> Self:
        self.fellows.add(self.symbol)
        return self


def representent_of(merge_history: SymbolMergeHistory[S]) -> SymbolMergeHistory[S]:
    if merge_history.representent is None:
        return merge_history
    return representent_of(merge_history.representent)


def merge_symbols(
    symbol_history_to_merge: SymbolMergeHistory[S],
    symbols_histories: list[SymbolMergeHistory[S]],
    all_symbols: list[S],
) -> None:
    """Merge all nodes."""
    for other_history in symbols_histories:
        _merge_one(symbol_history_to_merge, other_history)

    for symbol_history in symbols_histories + [symbol_history_to_merge]:
        if representent_of(symbol_history) != symbol_history:
            all_symbols.remove(symbol_history.symbol)


def _merge_one(
    symbol: SymbolMergeHistory[S], other_symbol: SymbolMergeHistory[S]
) -> SymbolMergeHistory[S]:
    self_rep = representent_of(symbol)
    other_rep = representent_of(other_symbol)
    if self_rep == other_rep:
        return self_rep

    other_rep.representent = self_rep
    self_rep.fellows.update(other_rep.fellows)
    other_rep.fellows.update(symbol.fellows)
    return self_rep



from __future__ import annotations

import itertools
from typing import Literal

from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import NumericalPredicate
from newclid.symbols.points_registry import Point


class Diff(NumericalPredicate):
    """diff a b -

    Represent that a is not equal to b.

    Numerical only.
    """

    predicate_type: Literal[PredicateType.DIFFERENT] = PredicateType.DIFFERENT
    points: tuple[Point, ...]

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        return tuple(sorted(args))

    def check_numerical(self) -> bool:
        existing_points = [p for p in self.points]
        for p, q in itertools.combinations(existing_points, 2):
            if p.num.close_enough(q.num):
                return False
        return True

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in self.points)

    def __str__(self) -> str:
        if len(self.points) == 2:
            return f"{self.points[0]} ≠ {self.points[1]}"
        else:
            return f"{', '.join(str(p) for p in self.points)} are all distinct"



from __future__ import annotations

from enum import Enum
from typing import Any, Iterable, cast

import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from newclid.animation.artists_in_animation import init_point_name_to_artist
from newclid.animation.clause import ClauseArtists, init_clauses_to_draw
from newclid.animation.goal import init_goals_to_draw
from newclid.animation.justifications import init_justifications_to_draw
from newclid.animation.text_animation import TextAnimation
from newclid.draw.figure import init_figure
from newclid.draw.theme import DrawTheme
from newclid.jgex.clause import JGEXClause
from newclid.jgex.formulation import JGEXFormulation
from newclid.justifications._index import JustificationType
from newclid.predicate_types import PredicateArgument
from newclid.proof_data import ProofData
from newclid.symbols.symbols_registry import SymbolsRegistry


class ProofAnimationParts(Enum):
    SHOWCASE = "Showcase of the final figure"
    CONSTRUCTION = "Construction of the initial figure"
    PROOF = "Proof"
    GOAL = "Summary of proven goals"


class ProofAnimation:
    def __init__(
        self,
        proof_data: ProofData,
        symbols: SymbolsRegistry,
        theme: DrawTheme,
        jgex_problem: JGEXFormulation | None = None,
        figure_kwargs: dict[str, Any] | None = None,
    ):
        self.jgex_problem = jgex_problem
        self.symbols = symbols
        self.theme = theme

        figure_kwargs = figure_kwargs or {"figsize": (24, 12), "dpi": 80}
        self.fig, axes = cast(
            tuple[Figure, dict[str, Axes]],
            plt.subplot_mosaic(  # pyright: ignore
                [
                    ["figure", "figure", "text", "text"],
                    ["figure", "figure", "text", "text"],
                ],
                **figure_kwargs,
            ),
        )

        self.figure_ax = axes["figure"]
        self.text_ax = axes["text"]

        init_figure(self.figure_ax, list(self.symbols.points))

        self.aux_point_names: set[PredicateArgument] = set()
        if self.jgex_problem is not None:
            self.aux_point_names = set(
                pname
                for clause in self.jgex_problem.auxiliary_clauses
                for pname in clause.points
            )

        self.point_name_to_artist = init_point_name_to_artist(
            ax=self.figure_ax,
            color_theme=self.theme,
            aux_point_names=self.aux_point_names,
            symbols=self.symbols,
        )

        self.goals_artists = init_goals_to_draw(
            ax=self.figure_ax,
            proven_goals=proof_data.proven_goals,
            point_name_to_artist=self.point_name_to_artist,
            symbols=self.symbols,
            theme=self.theme,
        )

        self.artists_by_clause: dict[JGEXClause, ClauseArtists] = {}
        self.clauses_to_draw: list[JGEXClause] = []
        if self.jgex_problem is not None:
            self.artists_by_clause = init_clauses_to_draw(
                ax=self.figure_ax,
                jgex_problem=self.jgex_problem,
                point_name_to_artist=self.point_name_to_artist,
                symbols=self.symbols,
                theme=self.theme,
            )
            self.clauses_to_draw = list(self.artists_by_clause.keys())

        (self.justifications_to_draw, self.justification_artists) = (
            init_justifications_to_draw(
                ax=self.figure_ax,
                point_name_to_artist=self.point_name_to_artist,
                symbols=self.symbols,
                proof_data=proof_data,
                theme=self.theme,
            )
        )

        text_lines: list[str] = [ProofAnimationParts.SHOWCASE.value]
        if self.jgex_problem is not None:
            text_lines.extend(
                [ProofAnimationParts.CONSTRUCTION.value]
                + [str(clause) for clause in self.clauses_to_draw]
            )

        text_lines.extend(
            [ProofAnimationParts.PROOF.value]
            + [
                justification_to_draw.text
                for justification_to_draw in self.justifications_to_draw
            ]
        )
        text_lines.extend([ProofAnimationParts.GOAL.value])

        self.text_animation = TextAnimation(
            self.text_ax,
            text_lines,
            theme=self.theme,
            n_lines_shown=21,
            fontsize=18,
        )
        self.text_animation.init()

        self.current_part = ProofAnimationParts.SHOWCASE
        self._has_done_showcase_transition: bool = False
        self._has_done_construction_transition: bool = False
        self._has_done_proof_transition: bool = False

        problem_name = (
            self.jgex_problem.name if self.jgex_problem is not None else "Problem"
        )
        plt.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
        self.fig.suptitle(  # pyright: ignore
            f"Newclid solution for {problem_name} | Goals: {', '.join(str(goal_in_proof.predicate) for goal_in_proof in proof_data.proven_goals)}",
            fontsize=30,
            # fontweight="bold",
            color=self.theme.title_color,
            family=["DejaVu Sans", "STIXGeneral"],
        )
        self.title = self.figure_ax.text(  # pyright: ignore
            0.01,
            0.01,
            self.current_part.value,
            transform=self.figure_ax.transAxes,
            fontsize=22,
            fontweight="bold",
            ha="left",
            va="bottom",
            color=self.theme.title_color,
            family=["DejaVu Sans", "STIXGeneral"],
        )

    def animate(self) -> FuncAnimation:
        return FuncAnimation(
            self.fig,
            self._next_step,
            init_func=self._init,
            frames=len(self.text_animation.proof_lines) + 1,
            interval=400,
            blit=True,
        )

    def _init(self) -> Iterable[Artist]:
        self.current_part = ProofAnimationParts.SHOWCASE
        return set()

    def _next_step(self, frame: int) -> Iterable[Artist]:
        match self.current_part:
            case ProofAnimationParts.SHOWCASE:
                return self._showcase(frame)
            case ProofAnimationParts.CONSTRUCTION:
                return self._construction(frame)
            case ProofAnimationParts.PROOF:
                return self._proof(frame)
            case ProofAnimationParts.GOAL:
                return self._goals(frame)

    def _showcase(self, frame: int) -> set[Artist]:
        if self._has_done_showcase_transition:
            return set()

        updated_artists: set[Artist] = set()
        updated_artists.update(self.text_animation.update(frame))
        updated_artists.update(self._hide_all_rule_applications())
        updated_artists.update(self._hide_all_extra_applications())
        updated_artists.update(self._hide_goals())
        self._has_done_showcase_transition = True
        if self.jgex_problem is not None:
            self.current_part = ProofAnimationParts.CONSTRUCTION
        else:
            self.current_part = ProofAnimationParts.PROOF
        return updated_artists

    def _construction(self, frame: int) -> set[Artist]:
        updated_artists: set[Artist] = set()
        if not self._has_done_construction_transition:
            self.title.set_text(ProofAnimationParts.CONSTRUCTION.value)
            updated_artists.add(self.title)
            updated_artists.update(self.text_animation.update(frame))
            for _pname, point_artists in self.point_name_to_artist.items():
                updated_artists.update(point_artists.hide())
            for clause, clause_artists in self.artists_by_clause.items():
                updated_artists.update(clause_artists.hide())

            self._has_done_construction_transition = True
            return updated_artists

        if not self.clauses_to_draw:
            raise ValueError("Should have done transition to proof")
        clause = self.clauses_to_draw.pop(0)
        if not self.clauses_to_draw:
            self.current_part = ProofAnimationParts.PROOF

        for other_clause, other_clause_artists in self.artists_by_clause.items():
            if other_clause == clause:
                continue
            updated_artists.update(other_clause_artists.gray_out())

        updated_artists.update(self.text_animation.update(frame))
        clause_artists = self.artists_by_clause[clause]
        if self.jgex_problem is None:
            raise ValueError("Should have a jgex problem to draw construction")

        is_aux = clause in self.jgex_problem.auxiliary_clauses
        highlight_color = (
            self.theme.aux_point_color if is_aux else self.theme.construction_color
        )
        updated_artists.update(
            clause_artists.manifest(
                aux_point_names=self.aux_point_names, highlight_color=highlight_color
            )
        )
        return updated_artists

    def _proof(self, frame: int) -> set[Artist]:
        updated_artists: set[Artist] = set()
        if not self._has_done_proof_transition:
            self.title.set_text(ProofAnimationParts.PROOF.value)
            updated_artists.add(self.title)
            updated_artists.update(self.text_animation.update(frame))
            updated_artists.update(self._gray_out_all_constructions())
            updated_artists.update(self._highlight_goals())
            self._has_done_proof_transition = True
            return updated_artists

        if not self.justifications_to_draw:
            raise ValueError("Should have done transition to goals")

        updated_artists.update(self._gray_out_all_constructions())
        updated_artists.update(self._gray_out_all_rule_applications())
        updated_artists.update(self._hide_goals())
        updated_artists.update(self._hide_all_extra_applications())

        next_justification = self.justifications_to_draw.pop(0)
        if not self.justifications_to_draw:
            self.current_part = ProofAnimationParts.GOAL

        self.text_animation.update(frame)
        justification_artists = self.justification_artists[
            next_justification.justification
        ]
        updated_artists.update(
            justification_artists.highlight(aux_point_names=self.aux_point_names)
        )
        return updated_artists

    def _goals(self, frame: int) -> set[Artist]:
        updated_artists: set[Artist] = set()
        updated_artists.update(self._gray_out_all_constructions())
        updated_artists.update(self._gray_out_all_rule_applications())
        updated_artists.update(self._hide_all_extra_applications())
        updated_artists.update(self.text_animation.update(frame))
        updated_artists.update(self._highlight_goals())
        return updated_artists

    def _gray_out_all_constructions(self) -> set[Artist]:
        updated_artists: set[Artist] = set()
        for clause_artists in self.artists_by_clause.values():
            updated_artists.update(clause_artists.gray_out())
        return updated_artists

    def _gray_out_all_rule_applications(self) -> set[Artist]:
        updated_artists: set[Artist] = set()
        for _, rule_application in self.justification_artists.items():
            if (
                rule_application.justification.dependency_type
                == JustificationType.RULE_APPLICATION
            ):
                updated_artists.update(rule_application.gray_out())
        return updated_artists

    def _hide_goals(self) -> set[Artist]:
        updated_artists: set[Artist] = set()
        for _, goal_artists in self.goals_artists.items():
            updated_artists.update(goal_artists.hide())
        return updated_artists

    def _highlight_goals(self) -> set[Artist]:
        updated_artists: set[Artist] = set()
        for _, goal_artists in self.goals_artists.items():
            updated_artists.update(
                goal_artists.highlight(
                    aux_point_names=self.aux_point_names,
                    highlight_color=self.theme.goal_color,
                )
            )
        return updated_artists

    def _hide_all_rule_applications(self) -> set[Artist]:
        updated_artists: set[Artist] = set()
        for _, rule_application in self.justification_artists.items():
            if (
                rule_application.justification.dependency_type
                == JustificationType.RULE_APPLICATION
            ):
                updated_artists.update(rule_application.hide())
        return updated_artists

    def _hide_all_extra_applications(self) -> set[Artist]:
        updated_artists: set[Artist] = set()
        for _dep, extra_application in self.justification_artists.items():
            if (
                extra_application.justification.dependency_type
                != JustificationType.RULE_APPLICATION
            ):
                updated_artists.update(extra_application.hide())
        return updated_artists



from enum import Enum


class PredicateType(str, Enum):
    DIFFERENT = "diff"
    MIDPOINT = "midp"
    COLLINEAR = "coll"
    N_COLLINEAR = "ncoll"

    PERPENDICULAR = "perp"
    N_PERPENDICULAR = "nperp"

    PARALLEL = "para"
    N_PARALLEL = "npara"

    CIRCUMCENTER = "circle"
    CYCLIC = "cyclic"

    CONSTANT_ANGLE = "aconst"
    A_COMPUTE = "acompute"
    OBTUSE_ANGLE = "obtuse_angle"

    EQUAL_ANGLES = "eqangle"

    CONSTANT_LENGTH = "lconst"
    L_COMPUTE = "lcompute"
    SQUARED_CONSTANT_LENGTH = "l2const"

    CONGRUENT = "cong"
    CONSTANT_RATIO = "rconst"
    R_COMPUTE = "rcompute"
    SQUARED_CONSTANT_RATIO = "r2const"

    EQUAL_RATIOS = "eqratio"

    SIMTRI_CLOCK = "simtri"
    SIMTRI_REFLECT = "simtrir"

    CONTRI_CLOCK = "contri"
    CONTRI_REFLECT = "contrir"

    SAME_CLOCK = "sameclock"
    SAME_SIDE = "sameside"
    N_SAME_SIDE = "nsameside"

    PYTHAGOREAN_PREMISES = "pythagorean_premises"
    PYTHAGOREAN_CONCLUSIONS = "pythagorean_conclusions"

    LENGTH_EQUATION = "lequation"
    ANGLE_EQUATION = "aequation"



from __future__ import annotations

from typing import Literal

from newclid.numerical import close_enough
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import PredicateInterface
from newclid.symbols.points_registry import Ratio


class EqRatio(PredicateInterface):
    """eqratio AB CD EF GH -

    Represent that AB/CD=EF/GH, as ratios between lengths of segments.
    """

    predicate_type: Literal[PredicateType.EQUAL_RATIOS] = PredicateType.EQUAL_RATIOS
    ratio1: Ratio
    ratio2: Ratio

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        a, b, c, d, e, f, g, h = args
        a, b = sorted((a, b))
        c, d = sorted((c, d))
        e, f = sorted((e, f))
        g, h = sorted((g, h))
        if min((a, b), (c, d)) > min((e, f), (g, h)):
            a, b, c, d, e, f, g, h = e, f, g, h, a, b, c, d
        if (a, b) > (c, d):
            a, b, c, d, e, f, g, h = c, d, a, b, g, h, e, f
        if (c, d) > (e, f):
            c, d, e, f = e, f, c, d
        return (a, b, c, d, e, f, g, h)

    def check_numerical(self) -> bool:
        (num, den) = (None, None)
        for (a, b), (c, d) in (self.ratio1, self.ratio2):
            (_num, _den) = (a.num.distance2(b.num), c.num.distance2(d.num))
            if num is not None and not close_enough(num * _den, _num * den):  # type: ignore
                return False
            num, den = _num, _den
        return True

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(
            PredicateArgument(p.name)
            for (a, b), (c, d) in (self.ratio1, self.ratio2)
            for p in (a, b, c, d)
        )

    def __str__(self) -> str:
        return " = ".join(
            f"{a}{b}:{c}{d}" for (a, b), (c, d) in (self.ratio1, self.ratio2)
        )



"""Utilities for Newclid."""

from __future__ import annotations

from difflib import Differ
from fractions import Fraction
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Optional, Sequence, TypeVar

from pydantic import BaseModel
from pyvis.network import Network  # type: ignore

from newclid.numerical import close_enough

if TYPE_CHECKING:
    from newclid.predicates import Predicate


class InfQuotientError(Exception):
    pass


# maximum denominator for a fraction.
MAX_DENOMINATOR = 1000000

T = TypeVar("T")


def get_quotient(v: Any) -> Fraction:
    v = float(v)
    n = v
    d = 1
    while not close_enough(n, round(n)):
        d += 1
        n += v
        if d > MAX_DENOMINATOR:
            e = InfQuotientError(v)
            raise e

    n = int(round(n))
    return Fraction(n, d)


def atomize(s: str, split_by: Optional[str] = None) -> tuple[str, ...]:
    words = s.split(split_by)
    return tuple(word.strip() for word in words)


def str_to_fraction(s: str) -> Fraction:
    if "pi/" in s:
        ns, ds = s.split("pi/")
        n, d = int(ns), int(ds)
        if d < 0:
            n, d = -n, -d
        return Fraction(n % d, d)
    elif "o" in s:
        n = int(s[:-1])
        d = 180
        return Fraction(n % d, d)
    elif "/" in s:
        ns, ds = s.split("/")
        n, d = int(ns), int(ds)
        return Fraction(n, d)
    else:
        n = int(s)
        d = 1
        return Fraction(n, d)


def fraction_to_len(f: Fraction):
    return f"{f.numerator}/{f.denominator}"


def fraction_to_ratio(f: Fraction):
    return f"{f.numerator}/{f.denominator}"


def fraction_to_angle(f: Fraction):
    n, d = f.numerator, f.denominator
    return f"{n % d}pi/{d}"


def reshape(to_reshape: Sequence[T], n: int) -> Generator[tuple[T, ...], None, None]:
    assert (len(to_reshape) % n) == 0, (
        f"Sequence length {len(to_reshape)} is not a multiple of reshape divisor {n}"
    )
    for i in range(0, len(to_reshape), n):
        yield tuple(to_reshape[i : i + n])


def add_edge(net: Network, u: Any, v: Any):
    net.add_node(u)  # type: ignore
    net.add_node(v)  # type: ignore
    net.add_edge(u, v)  # type: ignore


def run_static_server(directory_to_serve: Path):
    print(f"command to run the server: python -m http.server -d {directory_to_serve}")


def boring_predicate(predicate: Predicate) -> bool:
    s = str(predicate)
    if "=" in s:
        splited = atomize(s, "=")
        return all(t == splited[0] for t in splited)
    if "≅" in s:
        a, b = atomize(s, "≅")
        return a == b
    if "has the same orientation as" in s:
        a, b = atomize(s, "has the same orientation as")
        return a == b
    return False


def point_construction_tuple(point_name: str) -> tuple[str, ...]:
    return tuple(point_name[::-1]) if len(point_name) > 1 else ("0", *point_name[::-1])


S = TypeVar("S", bound=str)


def points_by_construction_order(points: set[S]) -> list[S]:
    return sorted(list(points), key=point_construction_tuple)


def pretty_basemodel_diff(actual: BaseModel, expected: BaseModel) -> str:
    return "\n".join(
        Differ().compare(
            expected.model_dump_json(indent=2).splitlines(),
            actual.model_dump_json(indent=2).splitlines(),
        )
    )


B = TypeVar("B", bound=BaseModel)


def pretty_basemodel_list_diff(actual: list[B], expected: list[B]) -> str:
    actual_combined_lines: list[str] = []
    for item in actual:
        actual_combined_lines.extend(item.model_dump_json(indent=2).splitlines())

    expected_combined_lines: list[str] = []
    for item in expected:
        expected_combined_lines.extend(item.model_dump_json(indent=2).splitlines())

    return "\n".join(Differ().compare(expected_combined_lines, actual_combined_lines))



# mypy: ignore-errors
from __future__ import annotations

import itertools
from fractions import Fraction
from typing import Any, Literal, Optional

from numpy import pi

from newclid.numerical import close_enough
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import PredicateInterface
from newclid.predicates.constant_angle import angle_between_4_points
from newclid.symbols.symbols_registry import SymbolsRegistry
from newclid.tools import fraction_to_angle, fraction_to_ratio, str_to_fraction


class AngleEquation(PredicateInterface):
    """aequation a A b B c C ... k -
    Represents the linear equation a * A + b * B + c * C + ... = k. The coefficients (small letters) are signed rational numbers,the capitalized letters are 4-tuples of points representing angles, k is an integer.

    The syntax of k is either a fraction of pi like 2pi/3 for radians or a number followed by a 'o' like 120o for degree.

    This generates an equation that can be added to the deductor for reasoning (AR for example).
    """

    predicate_type: Literal[PredicateType.ANGLE_EQUATION] = PredicateType.ANGLE_EQUATION
    coefficients: list[Fraction]
    angles: list[tuple[str, ...]]
    constant: Fraction

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        if (len(args) - 1) % 5 != 0:
            return None

        y = args[-1]
        pairs: list[tuple[Fraction, tuple[str, ...]]] = []
        for i in range(0, len(args) - 1, 5):
            coefficient = str_to_fraction(args[i])
            p1, p2, p3, p4 = args[i + 1 : i + 5]
            if p1 == p2 or p3 == p4:
                return None
            p1, p2 = sorted((p1, p2))
            p3, p4 = sorted((p3, p4))
            points = (p1, p2, p3, p4)
            pairs.append((coefficient, points))

        pairs_tuple = tuple(pairs)
        pairs_tuple = tuple(
            sorted(pairs_tuple, key=lambda pair: tuple(var for var in pair[1]))
        )

        # Merging coefficients of the same angle
        simplified_equation: list[tuple[Fraction, tuple[str, ...]]] = []
        for angle_points, group in itertools.groupby(
            pairs_tuple, key=lambda pair: pair[1]
        ):
            coefficient_sum = sum(pair[0] for pair in group)
            if coefficient_sum != 0:
                simplified_equation.append((coefficient_sum, angle_points))  # type: ignore

        if not simplified_equation:
            return None

        k = str_to_fraction(y)
        k %= 1
        equation: list[PredicateArgument] = [
            PredicateArgument(term)
            for coefficient, points in simplified_equation
            for term in (fraction_to_ratio(coefficient),) + points
        ]
        return tuple(equation + [PredicateArgument(fraction_to_angle(k))])

    @classmethod
    def parse(
        cls, args: tuple[PredicateArgument, ...], symbols: SymbolsRegistry
    ) -> Optional[tuple[Any, ...]]:
        preparse = cls.preparse(args)
        if not preparse:
            return None

        parsed: list[PredicateArgument] = []

        for i in range(0, len(preparse) - 1, 5):
            coefficient = str_to_fraction(preparse[i])
            points = preparse[i + 1 : i + 5]
            p1, p2, p3, p4 = symbols.points.names2points(points)
            parsed += [coefficient, p1, p2, p3, p4]  # type: ignore

        k = preparse[-1]
        return tuple(parsed + [str_to_fraction(k)])

    def check_numerical(self) -> bool:
        equation = self.args[:-1]
        expected_value = self.args[-1] * pi % pi
        evaluated_expression = 0

        for i in range(0, len(equation) - 1, 5):
            coefficient = float(equation[i])
            a, b, c, d = equation[i + 1 : i + 5]
            angle_value = angle_between_4_points(a, b, c, d)
            evaluated_expression += coefficient * angle_value

        evaluated_expression %= pi

        return close_enough(evaluated_expression, expected_value)

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        k = self.args[-1]
        equation = self.args[:-1]
        list_of_tokens: list[str] = []
        for i in range(0, len(equation) - 1, 5):
            coefficient = equation[i]
            p1, p2, p3, p4 = equation[i + 1 : i + 5]
            list_of_tokens.append(fraction_to_ratio(coefficient))
            list_of_tokens.append(p1.name)
            list_of_tokens.append(p2.name)
            list_of_tokens.append(p3.name)
            list_of_tokens.append(p4.name)
        list_of_tokens.append(fraction_to_angle(k))
        return tuple(list_of_tokens)

    def __str__(self) -> str:
        equation = self.args[0:-1]
        k = self.args[-1]
        pretty_string = ""

        for i in range(0, len(equation) - 1, 5):
            coefficient = equation[i]
            a, b, c, d = equation[i + 1 : i + 5]

            pretty_string += f"{fraction_to_ratio(coefficient)} ∠({a}{b},{c}{d}) + "

        pretty_string = pretty_string[:-2]  # Remove the last ' + '

        return pretty_string + f" = {fraction_to_angle(k)}"



from pathlib import Path
from typing import NamedTuple

import pytest
from newclid.ggb.problem_builder import GeogebraProblemBuilder
from newclid.numerical.geometries import PointNum
from newclid.problem import PredicateArgument, PredicateConstruction, ProblemSetup
from newclid.symbols.points_registry import Point
from newclid.tools import pretty_basemodel_diff


class GeogebraCase(NamedTuple):
    problem_id: str
    goals: list[tuple[str, ...]]
    expected_problem: ProblemSetup


def p(name: str) -> PredicateArgument:
    return PredicateArgument(name)


CASES = [
    GeogebraCase(
        problem_id="orthocenter",
        goals=[("perp", "A", "B", "C", "D")],
        expected_problem=ProblemSetup(
            name="orthocenter",
            goals=(PredicateConstruction.from_str("perp A B C D"),),
            points=(
                Point(name=p("A"), num=PointNum(x=0.0, y=1.0)),
                Point(name=p("B"), num=PointNum(x=5.0, y=8.0)),
                Point(name=p("C"), num=PointNum(x=8.0, y=3.0)),
                Point(
                    name=p("D"), num=PointNum(x=5.869565217391305, y=4.521739130434782)
                ),
            ),
            assumptions=(
                PredicateConstruction.from_str("perp B D A C"),
                PredicateConstruction.from_str("perp A D C B"),
            ),
        ),
    ),
    GeogebraCase(
        problem_id="indirect_parallelogram",
        goals=[("para", "A", "B", "C", "E")],
        expected_problem=ProblemSetup(
            name="indirect_parallelogram",
            goals=(PredicateConstruction.from_str(("para A B C E")),),
            points=(
                Point(name=p("A"), num=PointNum(x=-11.0, y=1.0)),
                Point(name=p("B"), num=PointNum(x=-8.0, y=4.0)),
                Point(name=p("C"), num=PointNum(x=-7.0, y=1.0)),
                Point(name=p("D"), num=PointNum(x=-14.0, y=4.0)),
                Point(name=p("E"), num=PointNum(x=-4.0, y=4.0)),
            ),
            assumptions=(
                PredicateConstruction.from_str(("para B E A C")),
                PredicateConstruction.from_str(("cong D A A B")),
                PredicateConstruction.from_str(("cong E C A D")),
            ),
        ),
    ),
    GeogebraCase(
        problem_id="circumcenter_midpoint",
        goals=[("eqangle", "A", "B", "A", "F", "A", "F", "A", "C")],
        expected_problem=ProblemSetup(
            name="circumcenter_midpoint",
            goals=(PredicateConstruction.from_str(("eqangle A B A F A F A C")),),
            points=(
                Point(name=p("A"), num=PointNum(x=0.0, y=0.0)),
                Point(name=p("B"), num=PointNum(x=6.0, y=-2.0)),
                Point(name=p("C"), num=PointNum(x=4.0, y=4.0)),
                Point(name=p("D"), num=PointNum(x=3.5, y=0.5)),
                Point(name=p("E"), num=PointNum(x=3.75, y=2.25)),
                Point(
                    name=p("F"), num=PointNum(x=3.23606797749979, y=0.7639320225002102)
                ),
                Point(
                    name=p("G"),
                    num=PointNum(x=2.6832815729997472, y=-0.8944271909999159),
                ),
                Point(name=p("H"), num=PointNum(x=2.0, y=2.0)),
            ),
            assumptions=(
                PredicateConstruction.from_str(("midp E D C")),
                PredicateConstruction.from_str(("coll A B G")),
                PredicateConstruction.from_str(("coll C A H")),
                PredicateConstruction.from_str(("eqangle C A A F A F A B")),
                PredicateConstruction.from_str(("eqangle B C B F B F A B")),
                PredicateConstruction.from_str(("perp F G A B")),
                PredicateConstruction.from_str(("circle D A C B")),
                PredicateConstruction.from_str(("cong H F F G")),
            ),
        ),
    ),
    GeogebraCase(
        problem_id="tangents",
        goals=[("cong", "D", "E", "D", "F")],
        expected_problem=ProblemSetup(
            name="tangents",
            goals=(PredicateConstruction.from_str(("cong D E D F")),),
            points=(
                Point(name=p("A"), num=PointNum(x=0.0, y=0.0)),
                Point(name=p("B"), num=PointNum(x=-10.0, y=0.0)),
                Point(name=p("C"), num=PointNum(x=-10.0, y=10.0)),
                Point(name=p("D"), num=PointNum(x=15.0, y=0.0)),
                Point(
                    name=p("E"), num=PointNum(x=6.666666666666667, y=7.453559924999299)
                ),
                Point(
                    name=p("F"), num=PointNum(x=6.666666666666667, y=-7.453559924999299)
                ),
            ),
            assumptions=(
                PredicateConstruction.from_str(("circle A B E F")),
                PredicateConstruction.from_str(("cong F A A B")),
                PredicateConstruction.from_str(("cong E A A B")),
                PredicateConstruction.from_str(("perp E A D E")),
                PredicateConstruction.from_str(("perp B A B C")),
                PredicateConstruction.from_str(("perp F A D F")),
            ),
        ),
    ),
    GeogebraCase(
        problem_id="imo_2019_p6",
        goals=[("perp", "A", "T", "A", "I")],
        expected_problem=ProblemSetup(
            name="imo_2019_p6",
            goals=(PredicateConstruction.from_str(("perp A T A I")),),
            points=(
                Point(name=p("A"), num=PointNum(x=-2.53, y=-0.6522616407982262)),
                Point(name=p("B"), num=PointNum(x=-0.19, y=2.09)),
                Point(name=p("C"), num=PointNum(x=1.49, y=-0.5322616407982262)),
                Point(
                    name=p("D"),
                    num=PointNum(x=0.5375646063301677, y=0.9543662152487727),
                ),
                Point(
                    name=p("E"),
                    num=PointNum(x=-0.27477205442495484, y=-0.5849414036168816),
                ),
                Point(
                    name=p("F"),
                    num=PointNum(x=-1.0654589167739486, y=1.064043843776083),
                ),
                Point(
                    name=p("I"),
                    num=PointNum(x=-0.3046153261720751, y=0.414808199911646),
                ),
                Point(
                    name=p("P"),
                    num=PointNum(x=0.4371408618192421, y=1.0857684432946726),
                ),
                Point(
                    name=p("Q"),
                    num=PointNum(x=-0.35280299034369944, y=0.5251464426438814),
                ),
                Point(
                    name=p("R"),
                    num=PointNum(x=-1.2526300070526193, y=0.09596959582020237),
                ),
                Point(
                    name=p("T"),
                    num=PointNum(x=-2.3984371937895226, y=-0.9266371703247638),
                ),
                Point(
                    name=p("O1"),
                    num=PointNum(x=0.5899354933306867, y=0.03362753959644931),
                ),
                Point(
                    name=p("O2"),
                    num=PointNum(x=-0.3175924089544161, y=1.3123787948883356),
                ),
            ),
            assumptions=(
                PredicateConstruction.from_str(("circle I D E F")),
                PredicateConstruction.from_str(("circle I D F R")),
                PredicateConstruction.from_str(("circle I D P R")),
                PredicateConstruction.from_str(("circle O1 C E P")),
                PredicateConstruction.from_str(("circle O1 C E Q")),
                PredicateConstruction.from_str(("circle O2 B F P")),
                PredicateConstruction.from_str(("circle O2 B F Q")),
                PredicateConstruction.from_str(("coll A B F")),
                PredicateConstruction.from_str(("coll A C E")),
                PredicateConstruction.from_str(("coll A P R")),
                PredicateConstruction.from_str(("coll B C D")),
                PredicateConstruction.from_str(("coll D I T")),
                PredicateConstruction.from_str(("coll P Q T")),
                PredicateConstruction.from_str(("cong D I E I")),
                PredicateConstruction.from_str(("cong E I I P")),
                PredicateConstruction.from_str(("cong F I E I")),
                PredicateConstruction.from_str(("cong I R E I")),
                PredicateConstruction.from_str(("cyclic B F P Q")),
                PredicateConstruction.from_str(("cyclic C E P Q")),
                PredicateConstruction.from_str(("cyclic D E F R")),
                PredicateConstruction.from_str(("cyclic D F P R")),
                PredicateConstruction.from_str(("eqangle A B A I A I A C")),
                PredicateConstruction.from_str(("eqangle A C C I C I B C")),
                PredicateConstruction.from_str(("perp A C E I")),
                PredicateConstruction.from_str(("perp D R E F")),
            ),
        ),
    ),
    GeogebraCase(
        problem_id="incenter",
        goals=[("eqangle", "A", "B", "B", "D", "B", "D", "B", "C")],
        expected_problem=ProblemSetup(
            name="incenter",
            goals=(PredicateConstruction.from_str(("eqangle A B B D B D B C")),),
            points=(
                Point(name=p("A"), num=PointNum(x=0.0, y=4.0)),
                Point(name=p("B"), num=PointNum(x=-4.0, y=0.0)),
                Point(name=p("C"), num=PointNum(x=4.0, y=0.0)),
                Point(name=p("D"), num=PointNum(x=0.0, y=1.6568542494923801)),
            ),
            assumptions=(
                PredicateConstruction.from_str(("eqangle A B B D B D B C")),
                PredicateConstruction.from_str(("eqangle A C C D C D B C")),
            ),
        ),
    ),
]


@pytest.mark.parametrize(
    "geogebra_case", CASES, ids=[case.problem_id for case in CASES]
)
def test_can_build_problem_from_geogebra(geogebra_case: GeogebraCase):
    ggb_file_path = Path(__file__).parent.joinpath(
        "ggb_exports", f"{geogebra_case.problem_id}.ggb"
    )
    ggb_problem_builder = GeogebraProblemBuilder(
        ggb_file_path=ggb_file_path
    ).with_goals(
        [PredicateConstruction.from_tuple(goal) for goal in geogebra_case.goals]
    )
    problem = ggb_problem_builder.build()
    expected = geogebra_case.expected_problem
    assert problem == expected, pretty_basemodel_diff(problem, expected)



from __future__ import annotations

from typing import Self

from pydantic import BaseModel, ConfigDict

from newclid.predicate_types import PredicateArgument
from newclid.tools import atomize, point_construction_tuple


class JGEXConstruction(BaseModel):
    model_config = ConfigDict(frozen=True)

    string: str
    """String representation of the construction."""

    @property
    def name(self) -> str:
        return self.string.split(" ")[0]

    @property
    def args(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(a) for a in self.string.split(" ")[1:])

    @classmethod
    def from_name_and_args(cls, name: str, args: tuple[PredicateArgument, ...]) -> Self:
        return cls.from_tuple((name, *args))

    @classmethod
    def from_str(cls, construction_str: str) -> Self:
        return cls(string=construction_str)

    @classmethod
    def from_tuple(cls, construction_tuple: tuple[str, ...]) -> Self:
        return cls(string=" ".join(construction_tuple))

    def __str__(self) -> str:
        return self.string

    def __hash__(self) -> int:
        return hash(self.string)


def rename_jgex_construction(
    construction: JGEXConstruction,
    mapping: dict[PredicateArgument, PredicateArgument],
) -> JGEXConstruction:
    renamed_args = tuple(mapping.get(arg, arg) for arg in construction.args)
    return JGEXConstruction.from_name_and_args(
        name=construction.name, args=renamed_args
    )


def is_numerical_argument(arg: str) -> bool:
    return not str.isalpha(arg[0])


class JGEXClause(BaseModel):
    model_config = ConfigDict(frozen=True)

    points: tuple[PredicateArgument, ...]
    constructions: tuple[JGEXConstruction, ...]

    def renamed(self, mp: dict[PredicateArgument, PredicateArgument]) -> JGEXClause:
        renamed_constructions: list[JGEXConstruction] = []
        for construction in self.constructions:
            renamed_constructions.append(rename_jgex_construction(construction, mp))
        return JGEXClause(
            points=tuple(mp.get(p, p) for p in self.points),
            constructions=tuple(renamed_constructions),
        )

    @classmethod
    def from_str(cls, s: str) -> tuple[JGEXClause, ...]:
        chunks = atomize(s, ";")

        clauses: list[JGEXClause] = []
        for chunk in chunks:
            points: tuple[PredicateArgument, ...] = tuple()
            points_str, construction_str = (
                chunk.split(":")
                if ":" in chunk
                else chunk.split("=")
                if "=" in chunk
                else ("", chunk)
            )
            points = tuple(PredicateArgument(p) for p in points_str.strip().split())
            construction_str = construction_str.strip()
            construction_tuples = tuple(
                atomize(b) for b in construction_str.split(",") if b.strip() != ""
            )
            clauses.append(
                JGEXClause(
                    points=points,
                    constructions=tuple(
                        JGEXConstruction.from_tuple(construction_tuple)
                        for construction_tuple in construction_tuples
                    ),
                )
            )
        return tuple(clauses)

    def __hash__(self) -> int:
        return hash((self.points, *self.constructions))

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return f"{' '.join(p for p in self.points)} = {', '.join(str(c) for c in self.constructions)}"


def order_clauses_by_points_construction_order(
    clauses: list[JGEXClause],
) -> list[JGEXClause]:
    """Order the clauses by the construction order of the points in the clause."""
    return sorted(
        clauses,
        key=lambda clause: min(point_construction_tuple(p) for p in clause.points),
    )



import numpy as np
import pytest
from newclid.numerical.geometries import PointNum
from newclid.predicates.collinearity import Coll
from newclid.predicates.cyclic import Cyclic
from newclid.symbols.circles_registry import CirclesRegistry
from newclid.symbols.lines_registry import LinesRegistry
from newclid.symbols.points_registry import Point


class TestLinesMerging:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.fixture = LinesRegistryFixture()

    def test_merge_two_lines_because_they_are_collinear(self):
        a, b, c = (
            Point(num=PointNum(x=0.0, y=0.0), name="a"),
            Point(num=PointNum(x=1.0, y=0.0), name="b"),
            Point(num=PointNum(x=2.0, y=0.0), name="c"),
        )
        self.fixture.given_registry_contains_line_thru_points({a, b})
        self.fixture.when_making_collinear(Coll(points=(a, b, c)))
        self.fixture.then_registry_should_contains_lines_thru_points([{a, b, c}])


class LinesRegistryFixture:
    def __init__(self):
        self.lines_registry = LinesRegistry()

    def given_registry_contains_line_thru_points(self, points: set[Point]):
        self.lines_registry.create_line_thru_points(points)

    def when_making_collinear(self, justification: Coll):
        self.lines_registry.make_collinear(
            points=justification.points, justification=justification
        )

    def then_registry_should_contains_lines_thru_points(self, lines: list[set[Point]]):
        expected_lines_str = [
            "-".join(sorted(point.name for point in line)) for line in lines
        ]
        assert len(self.lines_registry._lines) == len(lines), (  # pyright: ignore
            f"Registry contains lines {[str(line) for line in self.lines_registry]},"
            f" but expected only lines through points {expected_lines_str}"
        )
        for line in lines:
            assert self.lines_registry.line_containing(set(line)) is not None


class TestCirclesMerging:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.fixture = CirclesRegistryFixture()

    def test_merge_two_circles_because_they_are_cyclic(self):
        a, b, c, d, e = (
            Point(num=PointNum(x=0.0, y=0.0), name="a"),
            Point(num=PointNum(x=1.0, y=0.0), name="b"),
            Point(num=PointNum(x=0.0, y=1.0), name="c"),
            Point(num=PointNum(x=1.0, y=1.0), name="d"),
            Point(num=PointNum(x=np.cos(np.pi / 3), y=np.sin(np.pi / 3)), name="e"),
        )
        self.fixture.given_registry_contains_circle_thru_points(
            {a, b, c, d}, Cyclic(points=(a, b, c, d))
        )
        self.fixture.when_making_cyclic(Cyclic(points=(b, c, d, e)))
        self.fixture.then_registry_should_contains_circles_thru_points(
            [{a, b, c, d, e}]
        )


class CirclesRegistryFixture:
    def __init__(self):
        self.circles_registry = CirclesRegistry()

    def given_registry_contains_circle_thru_points(
        self, points: set[Point], because: Cyclic
    ):
        self.circles_registry.create_circle_thru_points(points, because=because)

    def when_making_cyclic(self, justification: Cyclic):
        self.circles_registry.make_cyclic(
            points=justification.points, justification=justification
        )

    def then_registry_should_contains_circles_thru_points(
        self, circles: list[set[Point]]
    ):
        expected_circles_str = [
            "-".join(sorted(point.name for point in circle)) for circle in circles
        ]
        assert len(self.circles_registry._circles) == len(circles), (  # pyright: ignore
            f"Registry contains circles {[str(circle) for circle in self.circles_registry]},"
            f" but expected only circles through points {expected_circles_str}"
        )
        for circle in circles:
            assert self.circles_registry.circle_containing(set(circle)) is not None



"""Tests manipulations of NcProblem."""

from typing import NamedTuple

import pytest
from newclid.numerical.geometries import PointNum
from newclid.predicate_types import PredicateArgument
from newclid.problem import (
    PredicateConstruction,
    ProblemSetup,
    filter_points_from_nc_problem,
    rename_points_in_nc_problem,
)
from newclid.symbols.points_registry import Point
from newclid.tools import pretty_basemodel_diff


def p(name: str) -> PredicateArgument:
    return PredicateArgument(name)


# A base problem for tests
BASE_PROBLEM = ProblemSetup(
    points=(
        Point(name=p("A"), num=PointNum(x=0, y=0)),
        Point(name=p("B"), num=PointNum(x=1, y=0)),
        Point(name=p("C"), num=PointNum(x=0, y=1)),
        Point(name=p("D"), num=PointNum(x=1, y=1)),
    ),
    assumptions=(
        PredicateConstruction.from_str("coll A B D"),
        PredicateConstruction.from_str("coll A B C D"),
    ),
    goals=(
        PredicateConstruction.from_str("coll B A C"),
        PredicateConstruction.from_str("coll B A D"),
    ),
)


class RenameCase(NamedTuple):
    name: str
    nc_problem: ProblemSetup
    mapping: dict[PredicateArgument, PredicateArgument]
    expected: ProblemSetup | Exception


RENAME_CASES = [
    RenameCase(
        name="Renaming one point",
        nc_problem=BASE_PROBLEM,
        mapping={p("A"): p("X")},
        expected=ProblemSetup(
            points=(
                Point(name=p("B"), num=PointNum(x=1, y=0)),
                Point(name=p("C"), num=PointNum(x=0, y=1)),
                Point(name=p("D"), num=PointNum(x=1, y=1)),
                Point(name=p("X"), num=PointNum(x=0, y=0)),
            ),
            assumptions=(
                PredicateConstruction.from_str("coll X B D"),
                PredicateConstruction.from_str("coll X B C D"),
            ),
            goals=(
                PredicateConstruction.from_str("coll B X C"),
                PredicateConstruction.from_str("coll B X D"),
            ),
        ),
    ),
    RenameCase(
        name="Renaming multiple points",
        nc_problem=BASE_PROBLEM,
        mapping={p("A"): p("X"), p("B"): p("Y")},
        expected=ProblemSetup(
            points=(
                Point(name=p("C"), num=PointNum(x=0, y=1)),
                Point(name=p("D"), num=PointNum(x=1, y=1)),
                Point(name=p("X"), num=PointNum(x=0, y=0)),
                Point(name=p("Y"), num=PointNum(x=1, y=0)),
            ),
            assumptions=(
                PredicateConstruction.from_str("coll X Y D"),
                PredicateConstruction.from_str("coll X Y C D"),
            ),
            goals=(
                PredicateConstruction.from_str("coll Y X C"),
                PredicateConstruction.from_str("coll Y X D"),
            ),
        ),
    ),
    RenameCase(
        name="Swapping two points",
        nc_problem=BASE_PROBLEM,
        mapping={p("A"): p("B"), p("B"): p("A")},
        expected=ProblemSetup(
            points=(
                Point(name=p("A"), num=PointNum(x=1, y=0)),
                Point(name=p("B"), num=PointNum(x=0, y=0)),
                Point(name=p("C"), num=PointNum(x=0, y=1)),
                Point(name=p("D"), num=PointNum(x=1, y=1)),
            ),
            assumptions=(
                PredicateConstruction.from_str("coll B A D"),
                PredicateConstruction.from_str("coll B A C D"),
            ),
            goals=(
                PredicateConstruction.from_str("coll A B C"),
                PredicateConstruction.from_str("coll A B D"),
            ),
        ),
    ),
    RenameCase(
        name="Empty mapping, no changes expected",
        nc_problem=BASE_PROBLEM,
        mapping={},
        expected=BASE_PROBLEM,
    ),
    RenameCase(
        name="Mapping a non-existent point, should raise ValueError",
        nc_problem=BASE_PROBLEM,
        mapping={p("Z"): p("X")},
        expected=ValueError(),
    ),
    RenameCase(
        name="Colliding a point to an existing point, should raise ValueError",
        nc_problem=BASE_PROBLEM,
        mapping={p("A"): p("B")},
        expected=ValueError(),
    ),
]


@pytest.mark.parametrize(
    "case",
    RENAME_CASES,
    ids=[c.name for c in RENAME_CASES],
)
def test_rename_points_in_nc_problem(case: RenameCase):
    """Test renaming points in an NcProblem."""
    if not isinstance(case.expected, ProblemSetup):
        with pytest.raises(case.expected.__class__):
            rename_points_in_nc_problem(case.nc_problem, case.mapping)
        return

    actual_problem = rename_points_in_nc_problem(case.nc_problem, case.mapping)
    assert actual_problem == case.expected, pretty_basemodel_diff(
        actual_problem, case.expected
    )


class FilterCase(NamedTuple):
    name: str
    nc_problem: ProblemSetup
    points_to_keep: list[PredicateArgument]
    expected: ProblemSetup | Exception


FILTER_CASES = [
    FilterCase(
        name="Keeping a subset of points",
        nc_problem=BASE_PROBLEM,
        points_to_keep=[p("A"), p("B"), p("D")],
        expected=ProblemSetup(
            points=(
                Point(name=p("A"), num=PointNum(x=0, y=0)),
                Point(name=p("B"), num=PointNum(x=1, y=0)),
                Point(name=p("D"), num=PointNum(x=1, y=1)),
            ),
            assumptions=(PredicateConstruction.from_str("coll A B D"),),
            goals=(PredicateConstruction.from_str("coll B A D"),),
        ),
    ),
    FilterCase(
        name="Keeping all points",
        nc_problem=BASE_PROBLEM,
        points_to_keep=[p("A"), p("B"), p("C"), p("D")],
        expected=BASE_PROBLEM,
    ),
    FilterCase(
        name="Keeping no points should raise ValueError",
        nc_problem=BASE_PROBLEM,
        points_to_keep=[],
        expected=ValueError(),
    ),
    FilterCase(
        name="Keeping points not in the problem should raise ValueError",
        nc_problem=BASE_PROBLEM,
        points_to_keep=[p("A"), p("Z")],
        expected=ValueError(),
    ),
]


@pytest.mark.parametrize(
    "case",
    FILTER_CASES,
    ids=[c.name for c in FILTER_CASES],
)
def test_filter_points_from_nc_problem(case: FilterCase):
    """Test filtering points from an NcProblem."""
    if not isinstance(case.expected, ProblemSetup):
        with pytest.raises(case.expected.__class__):
            filter_points_from_nc_problem(case.nc_problem, case.points_to_keep)
        return

    actual_problem = filter_points_from_nc_problem(case.nc_problem, case.points_to_keep)
    assert actual_problem == case.expected, pretty_basemodel_diff(
        actual_problem, case.expected
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Optional

from newclid.numerical.geometries import line_num_from_points
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import NumericalPredicate, PredicateInterface
from newclid.symbols.points_registry import Point

if TYPE_CHECKING:
    from newclid.proof_state import ProofState
    from newclid.symbols.symbols_registry import LineOrCircle, SymbolsRegistry


class Coll(PredicateInterface):
    """coll A B C ... -
    Represent that the 3 (or more) points in the arguments are collinear."""

    predicate_type: Literal[PredicateType.COLLINEAR] = PredicateType.COLLINEAR

    points: tuple[Point, ...]

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        if len(set(args)) <= 2:
            return None
        return tuple(sorted(set(args)))

    def check_numerical(self) -> bool:
        a, b = self.points[:2]
        line = line_num_from_points(a.num, b.num)
        return all(line.point_at(p.num.x, p.num.y) is not None for p in self.points[2:])

    def add(self, proof_state: ProofState) -> tuple[PredicateInterface, ...]:
        proof_state.symbols.lines.make_collinear(self.points, justification=self)
        return ()

    def check(self, proof_state: ProofState) -> bool | None:
        return proof_state.symbols.lines.check_collinear(self.points)

    def symbols(self, symbols: SymbolsRegistry) -> tuple[LineOrCircle, ...]:
        _rep, merged = symbols.lines.make_collinear(self.points, justification=self)
        return tuple(merged)

    def to_constructive(self, point: str) -> str:
        a, b, c = [p.name for p in self.points if p.name != point]
        if point == b:
            a, b = b, a
        if point == c:
            a, b, c = c, a, b
        return f"on_line {a} {b} {c}"

    def __str__(self) -> str:
        return f"{', '.join(str(p) for p in self.points)} are collinear"

    def __hash__(self) -> int:
        return hash(str(self))

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in self.points)


class NColl(NumericalPredicate):
    """ncoll A B C ... -
    Represent that any of the 3 (or mo}re) points is not aligned with the others.

    Numerical only.
    """

    predicate_type: Literal[PredicateType.N_COLLINEAR] = PredicateType.N_COLLINEAR

    points: tuple[Point, ...]

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> Optional[tuple[PredicateArgument, ...]]:
        return Coll.preparse(args)

    def check_numerical(self) -> bool:
        return not Coll(points=self.points).check_numerical()

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in self.points)

    def __str__(self) -> str:
        return f"{', '.join(str(p) for p in self.points)} are not collinear"



# mypy: disable-error-code="type-arg, misc"
# MyPy does a false positive on the generics here
# maybe it can be fixed but it's not a priority

from __future__ import annotations

import logging
from fractions import Fraction
from typing import Callable, Iterator, cast

from sympy import Expr  # type: ignore

from newclid.deductors import ARReason
from newclid.deductors.deductor_interface import (
    ARCoefficient,
    ARDeduction,
    ARPremise,
    Deductor,
)
from newclid.deductors.sympy_ar.ar_predicates import (
    aconst_expression,
    coll_expressions,
    cong_expressions,
    constant_length_expression,
    constant_ratio_expressions,
    enumerate_aconsts,
    enumerate_congs,
    enumerate_eqangles,
    enumerate_eqratios,
    enumerate_lconsts,
    enumerate_parallels,
    enumerate_perpendiculars,
    enumerate_rconsts,
    eqangle_expressions,
    eqratio_expressions,
    parallel_expressions,
    perpendicular_expressions,
)
from newclid.deductors.sympy_ar.ar_table import ARTable
from newclid.deductors.sympy_ar.table_angles import AnglesTable
from newclid.deductors.sympy_ar.table_ratios import RatiosTable
from newclid.justifications.justification import Justification
from newclid.predicates import Predicate, predicate_from_construction
from newclid.predicates._index import PredicateType
from newclid.problem import PredicateConstruction
from newclid.symbols.lines_registry import LineSymbol
from newclid.symbols.symbols_registry import SymbolsRegistry

LOGGER = logging.getLogger(__name__)


AngleEnumerator = Callable[
    [AnglesTable, SymbolsRegistry], Iterator[tuple[PredicateConstruction, Expr]]
]

RatioEnumerator = Callable[[RatiosTable], Iterator[tuple[PredicateConstruction, Expr]]]


class SympyARDeductor(Deductor):
    def __init__(self) -> None:
        self.atable = AnglesTable(ARTable())
        self.rtable = RatiosTable(ARTable())

        self.angle_enumerators: list[AngleEnumerator] = [
            enumerate_aconsts,
            enumerate_perpendiculars,
            enumerate_parallels,
            enumerate_eqangles,
        ]
        self.ratio_enumerators: list[RatioEnumerator] = [
            enumerate_lconsts,
            enumerate_congs,
            enumerate_rconsts,
            enumerate_eqratios,
        ]

        self._n_numericaly_false_deps_produced_by_ar = 0

    def deduce(self, symbols_registry: SymbolsRegistry) -> Iterator[Justification]:
        for angle_enumerator in self.angle_enumerators:
            for construction, _expression in angle_enumerator(
                self.atable, symbols_registry
            ):
                predicate = predicate_from_construction(
                    construction, symbols_registry.points
                )
                if predicate is None:
                    raise ValueError(f"AR built predicate {construction} is not valid")
                yield ARDeduction(predicate=predicate, ar_reason=ARReason.ANGLE_CHASING)

        for ratio_enumerator in self.ratio_enumerators:
            for construction, _expression in ratio_enumerator(self.rtable):
                predicate = predicate_from_construction(
                    construction, symbols_registry.points
                )
                if predicate is None:
                    raise ValueError(f"AR built predicate {construction} is not valid")
                yield ARDeduction(predicate=predicate, ar_reason=ARReason.RATIO_CHASING)

    def add_dependency(
        self, dependency: Justification, symbols_registry: SymbolsRegistry
    ) -> None:
        predicate = dependency.predicate
        table = self._table_for_predicate(predicate)
        if table is None:
            return
        expressions = predicate_to_expressions(predicate, symbols_registry, table)
        for expression in expressions:
            table.inner_table.add_expr(expression, predicate)

    def check_predicate(
        self, predicate: Predicate, symbols_registry: SymbolsRegistry
    ) -> bool:
        table = self._table_for_predicate(predicate)
        if table is None:
            return False
        expressions = predicate_to_expressions(predicate, symbols_registry, table)
        return all(table.inner_table.expr_delta(eq) for eq in expressions)

    def justify_predicate(
        self, predicate: Predicate, symbols_registry: SymbolsRegistry
    ) -> Justification:
        whys: list[Predicate] = []
        table = self._table_for_predicate(predicate)
        if table is None:
            raise ValueError(f"Cannot justify predicate {predicate} through AR")

        reason = reason_for_table(table)
        expressions = predicate_to_expressions(predicate, symbols_registry, table)
        for expression in expressions:
            whys.extend(table.inner_table.why_expr(expression))
        return ARDeduction(
            predicate=predicate,
            ar_reason=reason,
            ar_premises=tuple(
                ARPremise(
                    predicate=why,
                    coefficient=ARCoefficient(lhs_terms={}, coeff=Fraction(1)),
                )
                for why in whys
            ),
        )

    def _table_for_predicate(
        self, predicate: Predicate
    ) -> AnglesTable | RatiosTable | None:
        match predicate.predicate_type:
            # ANGLES
            case PredicateType.COLLINEAR:
                return self.atable
            case PredicateType.EQUAL_ANGLES:
                return self.atable
            case PredicateType.CONSTANT_ANGLE:
                return self.atable
            case PredicateType.PARALLEL:
                return self.atable
            case PredicateType.PERPENDICULAR:
                return self.atable
            # RATIOS
            case PredicateType.CONSTANT_LENGTH:
                return self.rtable
            case PredicateType.CONSTANT_RATIO:
                return self.rtable
            case PredicateType.CONGRUENT:
                return self.rtable
            case PredicateType.EQUAL_RATIOS:
                return self.rtable
            case _:
                return None


def reason_for_table(table: AnglesTable | RatiosTable) -> ARReason:
    if isinstance(table, AnglesTable):
        return ARReason.ANGLE_CHASING
    return ARReason.RATIO_CHASING


def predicate_to_expressions(
    predicate: Predicate,
    symbols_registry: SymbolsRegistry,
    table: AnglesTable | RatiosTable,
) -> list[Expr]:
    if isinstance(table, AnglesTable):
        return predicate_to_angle_expressions(predicate, symbols_registry, table)
    return predicate_to_ratio_expressions(predicate, table)


def predicate_to_angle_expressions(
    predicate: Predicate, symbols_registry: SymbolsRegistry, table: AnglesTable
) -> list[Expr]:
    match predicate.predicate_type:
        case PredicateType.COLLINEAR:
            lines = [
                cast(LineSymbol, symbol)
                for symbol in predicate.symbols(symbols_registry)
            ]
            return coll_expressions(lines, table=table)
        case PredicateType.EQUAL_ANGLES:
            l1, l2, l3, l4 = [
                cast(LineSymbol, symbol)
                for symbol in predicate.symbols(symbols_registry)
            ]
            return eqangle_expressions(l1, l2, l3, l4, table=table)
        case PredicateType.CONSTANT_ANGLE:
            l1, l2 = [
                cast(LineSymbol, symbol)
                for symbol in predicate.symbols(symbols_registry)
            ]
            return aconst_expression(l1, l2, predicate.angle, table=table)
        case PredicateType.PARALLEL:
            l1, l2 = [
                cast(LineSymbol, symbol)
                for symbol in predicate.symbols(symbols_registry)
            ]
            return parallel_expressions(l1, l2, table=table)
        case PredicateType.PERPENDICULAR:
            l1, l2 = [
                cast(LineSymbol, symbol)
                for symbol in predicate.symbols(symbols_registry)
            ]
            return perpendicular_expressions(l1, l2, table=table)
        case _:
            return []


def predicate_to_ratio_expressions(
    predicate: Predicate, table: RatiosTable
) -> list[Expr]:
    match predicate.predicate_type:
        case PredicateType.CONSTANT_LENGTH:
            return constant_length_expression(
                predicate.segment,
                predicate.length,
                table=table,
            )
        case PredicateType.CONSTANT_RATIO:
            return constant_ratio_expressions(
                predicate.ratio,
                predicate.value,
                table=table,
            )
        case PredicateType.CONGRUENT:
            return cong_expressions(
                predicate.segment1,
                predicate.segment2,
                table=table,
            )
        case PredicateType.EQUAL_RATIOS:
            return eqratio_expressions(
                predicate.ratio1,
                predicate.ratio2,
                table=table,
            )
        case _:
            return []



"""Random number generation."""

import numpy as np


def setup_rng(rng: np.random.Generator | int | None) -> np.random.Generator:
    if rng is None:
        return np.random.default_rng()
    if isinstance(rng, int):
        return np.random.default_rng(rng)
    return rng



import itertools

from newclid.rule_matching.permutations import generate_permutations_as_dicts


def test_output_is_correct():
    variables = ["a"]
    points = ["x", "y", "z", "w"]

    baseline_list = [
        {v: p for v, p in zip(variables, point_list)}
        for point_list in itertools.permutations(points, r=len(variables))
    ]

    permutation_generator_list = list(generate_permutations_as_dicts(points, variables))

    print(len(baseline_list))
    print(baseline_list[:3])
    print(permutation_generator_list[:3])

    assert len(baseline_list) == len(permutation_generator_list)

    baseline = set(tuple(sorted(mapping.items())) for mapping in baseline_list)
    permutation_generator_set = set(
        tuple(sorted(mapping.items())) for mapping in permutation_generator_list
    )

    assert baseline == permutation_generator_set



from __future__ import annotations

from typing import Literal

from newclid.numerical import sign
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import NumericalPredicate
from newclid.symbols.points_registry import Triangle


class SameSide(NumericalPredicate):
    """sameside a b c x y z

    Represent that a in the same kind of arc (smaller or larger) with respect to b-c as x is with respect to y-z (includes the degenerate case of the circle being a line).

    If a, b, c or x, y, z are not in the same circle/line, turns true if a is closer/further/at the same distance to the midpoint of b-c and, respectively, x is closer/further/at the same distance to the midpoint of y-z.

    Numerical only.
    """

    predicate_type: Literal[PredicateType.SAME_SIDE] = PredicateType.SAME_SIDE
    triangle1: Triangle
    triangle2: Triangle

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        a, b, c, x, y, z = args
        if a == b or b == c or a == c or x == y or y == z or x == z:
            return None
        p1 = min((a, b, c), (a, c, b))
        p2 = min((x, y, z), (x, z, y))
        return min(p1 + p2, p2 + p1)

    def check_numerical(self) -> bool:
        a, b, c = self.triangle1
        x, y, z = self.triangle2
        sa = sign((b.num - a.num).dot(c.num - a.num))
        sx = sign((y.num - x.num).dot(z.num - x.num))
        return sa == sx

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in self.triangle1 + self.triangle2)

    def __str__(self) -> str:
        a, b, c = self.triangle1
        x, y, z = self.triangle2
        return f"{a} is to the same side of {b}->{c} as {x} is to {y}->{z}"


class NSameSide(NumericalPredicate):
    """nsameside a b c x y z

    Represent that a is to the different side of b->c as x is to y->z.

    Numerical only.
    """

    predicate_type: Literal[PredicateType.N_SAME_SIDE] = PredicateType.N_SAME_SIDE
    triangle1: Triangle
    triangle2: Triangle

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        return SameSide.preparse(args)

    def check_numerical(self) -> bool:
        return not SameSide(
            triangle1=self.triangle1, triangle2=self.triangle2
        ).check_numerical()

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in self.triangle1 + self.triangle2)

    def __str__(self) -> str:
        a, b, c = self.triangle1
        x, y, z = self.triangle2
        return f"{a} is to the different side of {b}->{c} as {x} is to {y}->{z}"



"""Interface and implementations of Deductive Agents."""

from typing import Annotated

from pydantic import Field

from newclid.agent.ddarn import DDARNStats
from newclid.agent.follow_deductions import FollowDeductionsStats

AgentStats = Annotated[
    DDARNStats | FollowDeductionsStats, Field(discriminator="agent_type")
]



from pathlib import Path

import numpy as np
import pytest
from newclid.animation import html_animation
from newclid.api import GeometricSolverBuilder
from newclid.draw.theme import DrawTheme
from newclid.jgex.problem_builder import JGEXProblemBuilder


def test_draw_initial_orthocenter(tmp_path: Path):
    rng = np.random.default_rng(1234)
    jgex_problem_builder = JGEXProblemBuilder(rng=rng).with_problem_from_txt(
        "a b c = triangle a b c; h = on_tline h b a c, on_tline h c a b ? perp a h b c"
    )
    solver = GeometricSolverBuilder(rng=rng).build(jgex_problem_builder.build())

    out_dir = tmp_path / "orthocenter"

    # Draw the initial figure
    solver.draw_figure(
        out_file=out_dir / "initial_figure.svg",
        jgex_problem=jgex_problem_builder.jgex_problem,
    )

    # Run the solver
    solver.run()

    # Draw the final figure
    solver.draw_figure(
        out_file=out_dir / "final_figure.svg",
        jgex_problem=jgex_problem_builder.jgex_problem,
    )


@pytest.mark.skip(reason="Too slow to run")
def test_draw_animation_with_jgex_problem():
    rng = np.random.default_rng(0)

    jgex_problem_builder = (
        JGEXProblemBuilder(rng=rng)
        .include_auxiliary_clauses()
        .with_problem_from_txt(
            "a m n = triangle a m n"
            "; b = on_circle b m a, on_circle b n a"
            "; c = on_line c m n, on_circle c m a"
            "; d = on_line d m n, on_circle d n a"
            "; p = circle p a c d"
            "; e = on_line e a p, on_circle e m a"
            "; f = on_line f a p, on_circle f n a"
            "; h = on_tline h p m n, on_tline h m p n"
            "; o = circle o b e f"
            "; x = on_dia x o h, on_pline x h a p"
            " | w = circle w d e f"
            " ? cong o x o b"
            "; obtuse_angle c m n"
            "; obtuse_angle m n d"
        )
    )

    solver = GeometricSolverBuilder(rng=rng).build(jgex_problem_builder.build())
    solver.run()
    assert jgex_problem_builder.jgex_problem is not None
    custom_theme = DrawTheme(
        circle_color="#000000",
        line_color="#000000",
        triangle_color="#000000",
        goal_color="#000000",
        aux_point_color="#000000",
        construction_color="#000000",
        point_color="#000000",
        text_color="#000000",
        perpendicular_color="#000000",
        title_color="#000000",
        thick_line_width=10,
        thin_line_width=5,
    )
    html_animation(
        solver.animate(jgex_problem_builder.jgex_problem, theme=custom_theme)
    )
